#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import glob as globlib
import json
import logging
import math
import os
import re
import sys
import time
from json.decoder import JSONDecodeError

import click
import numpy as np
import pandas as pd
import requests
from bravado.exception import (
    BravadoConnectionError, BravadoTimeoutError, HTTPBadGateway, HTTPBadRequest,
    HTTPForbidden, HTTPGatewayTimeout, HTTPInternalServerError, HTTPRequestTimeout,
    HTTPServerError, HTTPServiceUnavailable, HTTPUnauthorized
)

from neptune import envs
from neptune.api_exceptions import ConnectionLost, Forbidden, SSLError, ServerError, Unauthorized
from neptune.exceptions import (
    FileNotFound, InvalidNotebookPath, NeptuneIncorrectProjectQualifiedNameException,
    NeptuneMissingProjectQualifiedNameException,
    NotADirectory, NotAFile, ProjectMigratedToNewStructure
)
from neptune.git_info import GitInfo
from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN

_logger = logging.getLogger(__name__)

IS_WINDOWS = sys.platform == 'win32'
IS_MACOS = sys.platform == 'darwin'

MIGRATION_IN_PROGRESS = 'PROJECT_MIGRATION_IN_PROGRESS'
MIGRATION_FINISHED = 'NEW_PROJECT_VERSION'


def map_values(f_value, dictionary):
    return dict(
        (k, f_value(v)) for k, v in dictionary.items()
    )


def map_keys(f_key, dictionary):
    return dict(
        (f_key(k), v) for k, v in dictionary.items()
    )


def as_list(value):
    if value is None or isinstance(value, list):
        return value
    else:
        return [value]


def validate_notebook_path(path):
    if not path.endswith(".ipynb"):
        raise InvalidNotebookPath(path)

    if not os.path.exists(path):
        raise FileNotFound(path)

    if not os.path.isfile(path):
        raise NotAFile(path)


def assure_directory_exists(destination_dir):
    """Check if `destination_dir` DIRECTORY exists, or creates one"""
    if not destination_dir:
        destination_dir = os.getcwd()

    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    elif not os.path.isdir(destination_dir):
        raise NotADirectory(destination_dir)

    return destination_dir


def align_channels_on_x(dataframe):
    channel_dfs, common_x = _split_df_by_stems(dataframe)
    return merge_dataframes([common_x] + channel_dfs, on='x', how='outer')


def get_channel_name_stems(columns):
    return list(set([col[2:] for col in columns]))


def merge_dataframes(dataframes, on, how='outer'):
    merged_df = functools.reduce(lambda left, right: \
                                     pd.merge(left, right, on=on, how=how), dataframes)
    return merged_df


def is_float(value):
    try:
        _ = float(value)
    except ValueError:
        return False
    else:
        return True


def is_nan_or_inf(value):
    return math.isnan(value) or math.isinf(value)


def file_contains(filename, text):
    with open(filename) as f:
        for line in f:
            if text in line:
                return True
        return False


def in_docker():
    cgroup_file = '/proc/self/cgroup'
    return os.path.exists('./dockerenv') or (os.path.exists(cgroup_file) and file_contains(cgroup_file, text='docker'))


def is_notebook():
    try:
        # pylint: disable=pointless-statement,undefined-variable
        get_ipython
        return True
    except Exception:
        return False


def _split_df_by_stems(df):
    channel_dfs, x_vals = [], []
    for stem in get_channel_name_stems(df.columns):
        channel_df = df[['x_{}'.format(stem), 'y_{}'.format(stem)]]
        channel_df.columns = ['x', stem]
        channel_df = channel_df.dropna()
        channel_dfs.append(channel_df)
        x_vals.extend(channel_df['x'].tolist())
    common_x = pd.DataFrame({'x': np.unique(x_vals)}, dtype=float)
    return channel_dfs, common_x


def discover_git_repo_location():
    # pylint:disable=bad-option-value,import-outside-toplevel
    import __main__

    if hasattr(__main__, '__file__'):
        return os.path.dirname(os.path.abspath(__main__.__file__))
    return None


def update_session_proxies(session, proxies):
    if proxies is not None:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def get_git_info(repo_path=None):
    """Retrieve information about git repository.

    If attempt fails, ``None`` will be returned.

    Args:
        repo_path (:obj:`str`, optional, default is ``None``):

            | Path to the repository from which extract information about git.
            | If ``None`` is passed, calling ``get_git_info`` is equivalent to calling
              ``git.Repo(search_parent_directories=True)``.
              Check `GitPython <https://gitpython.readthedocs.io/en/stable/reference.html#git.repo.base.Repo>`_
              docs for more information.

    Returns:
        :class:`~neptune.git_info.GitInfo` - An object representing information about git repository.

    Examples:

        .. code:: python3

            # Get git info from the current directory
            git_info = get_git_info('.')

    """
    try:
        # pylint:disable=bad-option-value,import-outside-toplevel
        import git

        repo = git.Repo(repo_path, search_parent_directories=True)

        commit = repo.head.commit

        active_branch = ""

        try:
            active_branch = repo.active_branch.name
        except TypeError as e:
            if str(e.args[0]).startswith("HEAD is a detached symbolic reference as it points to"):
                active_branch = "Detached HEAD"

        remote_urls = [remote.url for remote in repo.remotes]

        return GitInfo(
            commit_id=commit.hexsha,
            message=commit.message,
            author_name=commit.author.name,
            author_email=commit.author.email,
            commit_date=commit.committed_datetime,
            repository_dirty=repo.is_dirty(untracked_files=True),
            active_branch=active_branch,
            remote_urls=remote_urls
        )
    except:  # pylint: disable=bare-except
        return None


def parse_error_type(ex):
    try:
        error_data = json.loads(ex.response.text)
        return error_data.get("errorType")
    except JSONDecodeError:
        return None


migration_reported = False


def print_migration_in_progress_message():
    # pylint: disable=global-statement
    global migration_reported
    migration_reported = True
    click.echo(click.style("""
NOTICE: Attention needed
Your Neptune project is currently being migrated to the new structure. 
The execution of the script is paused until migration is finished. 
It will automatically resume once it's finished.
Depending on the number of runs (experiments) in the project
the migration process can take from few minutes to up to few hours.
If you think this operation takes too long please contact Neptune support:
    - https://neptune.ai/?chat-with-us
""", fg='yellow'))


def with_api_exceptions_handler(func):
    def wrapper(*args, **kwargs):
        # pylint: disable=global-statement
        global migration_reported
        retries = 11
        retry = 0
        while retry < retries:
            try:
                return func(*args, **kwargs)
            except requests.exceptions.SSLError:
                raise SSLError()
            except HTTPBadRequest as e:
                if parse_error_type(e) == MIGRATION_FINISHED:
                    raise ProjectMigratedToNewStructure()
                else:
                    raise
            except HTTPServiceUnavailable as e:
                if parse_error_type(e) == MIGRATION_IN_PROGRESS:
                    retry = min(retry + 1, 8)
                    if not migration_reported:
                        print_migration_in_progress_message()
                    time.sleep(2 ** retry)
                    continue
                else:
                    if retry >= 6:
                        _logger.warning(
                            'Experiencing connection interruptions. Reestablishing communication with Neptune.')
                    time.sleep(2 ** retry)
                    retry += 1
                    continue
            except (BravadoConnectionError, BravadoTimeoutError,
                    requests.exceptions.ConnectionError, requests.exceptions.Timeout,
                    HTTPRequestTimeout, HTTPGatewayTimeout, HTTPBadGateway):
                if retry >= 6:
                    _logger.warning('Experiencing connection interruptions. Reestablishing communication with Neptune.')
                time.sleep(2 ** retry)
                retry += 1
                continue
            except HTTPServerError:
                raise ServerError()
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code == HTTPBadRequest.status_code and parse_error_type(e) == MIGRATION_FINISHED:
                    raise ProjectMigratedToNewStructure()
                elif status_code == HTTPServiceUnavailable.status_code and parse_error_type(e) == MIGRATION_IN_PROGRESS:
                    retry = min(retry + 1, 8)
                    if not migration_reported:
                        print_migration_in_progress_message()
                    time.sleep(2 ** retry)
                    continue
                elif status_code in (
                        HTTPBadGateway.status_code,
                        HTTPServiceUnavailable.status_code,
                        HTTPGatewayTimeout.status_code):
                    if retry >= 6:
                        _logger.warning(
                            'Experiencing connection interruptions. Reestablishing communication with Neptune.')
                    time.sleep(2 ** retry)
                    retry += 1
                    continue
                elif status_code >= HTTPInternalServerError.status_code:
                    raise ServerError()
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                else:
                    raise
        raise ConnectionLost()

    return wrapper


def glob(pathname):
    # pylint: disable=unexpected-keyword-arg
    if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 5):
        return globlib.glob(pathname)
    else:
        return globlib.glob(pathname, recursive=True)


def is_ipython():
    try:
        # pylint:disable=bad-option-value,import-outside-toplevel
        import IPython
        ipython = IPython.core.getipython.get_ipython()
        return ipython is not None
    except ImportError:
        return False


def assure_project_qualified_name(project_qualified_name):
    project_qualified_name = project_qualified_name or os.getenv(envs.PROJECT_ENV_NAME)

    if not project_qualified_name:
        raise NeptuneMissingProjectQualifiedNameException()
    if not re.match(PROJECT_QUALIFIED_NAME_PATTERN, project_qualified_name):
        raise NeptuneIncorrectProjectQualifiedNameException(project_qualified_name)

    return project_qualified_name


class NoopObject(object):

    def __getattr__(self, name):
        return self

    def __getitem__(self, key):
        return self

    def __call__(self, *args, **kwargs):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc
from typing import Generic, TypeVar, List, Optional, Tuple

T = TypeVar('T')


class StorageQueue(Generic[T]):

    # NOTICE: All implementations should be thread-safe as long as there is only one consumer and one producer.

    @abc.abstractmethod
    def put(self, obj: T) -> int:
        pass

    @abc.abstractmethod
    def get(self) -> Tuple[Optional[T], int]:
        pass

    @abc.abstractmethod
    def get_batch(self, size: int) -> Tuple[Optional[List[T]], int]:
        pass

    @abc.abstractmethod
    def flush(self) -> None:
        pass

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def wait_for_empty(self, seconds: Optional[float] = None) -> None:
        pass

    @abc.abstractmethod
    def ack(self, version: int) -> None:
        pass

    @abc.abstractmethod
    def is_empty(self) -> bool:
        pass

    @abc.abstractmethod
    def size(self) -> int:
        pass



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

from mock import MagicMock

from neptune.new.internal.operation import AssignString
from neptune.new.attributes.atoms.string import String, StringVal

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestString(TestAttributeBase):

    def test_assign(self):
        value_and_expected = [
            ("qwertyu", "qwertyu"),
            (StringVal("Some string"), "Some string")
        ]

        for value, expected in value_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = String(exp, path)
            var.assign(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(AssignString(path, expected), wait)

    def test_get(self):
        exp, path = self._create_run(), self._random_path()
        var = String(exp, path)
        var.assign("adfh")
        self.assertEqual("adfh", var.fetch())



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import threading

from typing import TextIO

from neptune.new.run import Run
from neptune.new.logging import Logger as NeptuneLogger


class StdStreamCaptureLogger:

    def __init__(self, run: Run, attribute_name: str, stream: TextIO):
        self._logger = NeptuneLogger(run, attribute_name)
        self.stream = stream
        self._thread_local = threading.local()

    def write(self, data: str):
        if not hasattr(self._thread_local, "inside_write"):
            self._thread_local.inside_write = False

        self.stream.write(data)
        if not self._thread_local.inside_write:
            try:
                self._thread_local.inside_write = True
                self._logger.log(data)
            finally:
                self._thread_local.inside_write = False

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


class StdoutCaptureLogger(StdStreamCaptureLogger):

    def __init__(self, run: Run, attribute_name: str):
        super().__init__(run, attribute_name, sys.stdout)
        sys.stdout = self

    def close(self):
        sys.stdout = self.stream


class StderrCaptureLogger(StdStreamCaptureLogger):

    def __init__(self, run: Run, attribute_name: str):
        super().__init__(run, attribute_name, sys.stderr)
        sys.stderr = self

    def close(self):
        sys.stderr = self.stream



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import time
import traceback
import logging
from logging import StreamHandler

from neptune.internal.abort import DefaultAbortImpl, CustomAbortImpl
from neptune.internal.channels.channels import ChannelNamespace
from neptune.internal.hardware.gauges.gauge_mode import GaugeMode
from neptune.internal.hardware.metrics.service.metric_service_factory import MetricServiceFactory
from neptune.internal.hardware.system.system_monitor import SystemMonitor
from neptune.internal.streams.channel_writer import ChannelWriter
from neptune.internal.streams.stdstream_uploader import StdOutWithUpload, StdErrWithUpload
from neptune.internal.threads.aborting_thread import AbortingThread
from neptune.internal.threads.hardware_metric_reporting_thread import HardwareMetricReportingThread
from neptune.internal.threads.ping_thread import PingThread
from neptune.utils import is_notebook, in_docker, is_ipython

_logger = logging.getLogger(__name__)


class ExecutionContext(object):

    def __init__(self, backend, experiment):
        self._backend = backend
        self._experiment = experiment
        self._ping_thread = None
        self._hardware_metric_thread = None
        self._aborting_thread = None
        self._logger = None
        self._logger_handler = None
        self._stdout_uploader = None
        self._stderr_uploader = None
        self._uncaught_exception_handler = sys.__excepthook__

        self._previous_uncaught_exception_handler = None

    def start(self,
              abort_callback=None,
              logger=None,
              upload_stdout=True,
              upload_stderr=True,
              send_hardware_metrics=True,
              run_monitoring_thread=True,
              handle_uncaught_exceptions=True):

        if handle_uncaught_exceptions:
            self._set_uncaught_exception_handler()

        if logger:
            # pylint: disable=protected-access
            channel = self._experiment._get_channel('logger', 'text', ChannelNamespace.SYSTEM)
            channel_writer = ChannelWriter(self._experiment, channel.name, ChannelNamespace.SYSTEM)
            self._logger_handler = StreamHandler(channel_writer)
            self._logger = logger
            logger.addHandler(self._logger_handler)

        if upload_stdout and not is_notebook():
            self._stdout_uploader = StdOutWithUpload(self._experiment)

        if upload_stderr and not is_notebook():
            self._stderr_uploader = StdErrWithUpload(self._experiment)

        abortable = abort_callback is not None or DefaultAbortImpl.requirements_installed()
        if abortable:
            self._run_aborting_thread(abort_callback)
        else:
            _logger.warning('psutil is not installed. You will not be able to abort this experiment from the UI.')

        if run_monitoring_thread:
            self._run_monitoring_thread()

        if send_hardware_metrics:
            if SystemMonitor.requirements_installed():
                self._run_hardware_metrics_reporting_thread()
            else:
                _logger.warning('psutil is not installed. Hardware metrics will not be collected.')

    def stop(self):
        if self._ping_thread:
            self._ping_thread.interrupt()
            self._ping_thread = None

        if self._hardware_metric_thread:
            self._hardware_metric_thread.interrupt()
            self._hardware_metric_thread = None

        if self._aborting_thread:
            self._aborting_thread.shutdown()
            self._aborting_thread = None

        if self._stdout_uploader:
            self._stdout_uploader.close()

        if self._stderr_uploader:
            self._stderr_uploader.close()

        if self._logger and self._logger_handler:
            self._logger.removeHandler(self._logger_handler)

        sys.excepthook = self._previous_uncaught_exception_handler

    def _set_uncaught_exception_handler(self):

        def exception_handler(exc_type, exc_val, exc_tb):
            self._experiment.stop("\n".join(traceback.format_tb(exc_tb)) + "\n" + repr(exc_val))

            sys.__excepthook__(exc_type, exc_val, exc_tb)

        self._uncaught_exception_handler = exception_handler

        self._previous_uncaught_exception_handler = sys.excepthook
        sys.excepthook = exception_handler

    def _run_aborting_thread(self, abort_callback):
        if abort_callback is not None:
            abort_impl = CustomAbortImpl(abort_callback)
        elif not is_ipython():
            abort_impl = DefaultAbortImpl(pid=os.getpid())
        else:
            return

        websocket_factory = self._backend.websockets_factory(
            # pylint: disable=protected-access
            project_uuid=self._experiment._project.internal_id,
            experiment_id=self._experiment.internal_id
        )
        if not websocket_factory:
            return

        self._aborting_thread = AbortingThread(
            websocket_factory=websocket_factory,
            abort_impl=abort_impl,
            experiment=self._experiment
        )
        self._aborting_thread.start()

    def _run_monitoring_thread(self):
        self._ping_thread = PingThread(backend=self._backend, experiment=self._experiment)
        self._ping_thread.start()

    def _run_hardware_metrics_reporting_thread(self):
        gauge_mode = GaugeMode.CGROUP if in_docker() else GaugeMode.SYSTEM
        metric_service = MetricServiceFactory(self._backend, os.environ).create(
            gauge_mode=gauge_mode,
            experiment=self._experiment,
            reference_timestamp=time.time()
        )

        self._hardware_metric_thread = HardwareMetricReportingThread(
            metric_service=metric_service,
            metric_sending_interval_seconds=10
        )
        self._hardware_metric_thread.start()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import imghdr
import os
import pathlib

from typing import Optional, Iterable

from neptune.new.internal.utils import base64_encode

from neptune.new.exceptions import FileNotFound, OperationNotSupported

from neptune.new.types import File
from neptune.new.types.series.file_series import FileSeries as FileSeriesVal
from neptune.new.internal.operation import ImageValue, LogImages, ClearImageLog, Operation
from neptune.new.attributes.series.series import Series

Val = FileSeriesVal
Data = File


class FileSeries(Series[Val, Data]):

    def _get_log_operation_from_value(self, value: Val, step: Optional[float], timestamp: float) -> Operation:
        values = [
            LogImages.ValueType(
                ImageValue(data=self._get_base64_image_content(val), name=value.name, description=value.description),
                step=step,
                ts=timestamp)
            for val in value.values
        ]
        return LogImages(self._path, values)

    def _get_clear_operation(self) -> Operation:
        return ClearImageLog(self._path)

    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        return FileSeriesVal(values, **kwargs)

    def _is_value_type(self, value) -> bool:
        return isinstance(value, FileSeriesVal)

    @staticmethod
    def _get_base64_image_content(file: File) -> str:
        if file.path is not None:
            if not os.path.exists(file.path):
                raise FileNotFound(file.path)
            with open(file.path, 'rb') as image_file:
                file = File.from_stream(image_file)

        ext = imghdr.what("", h=file.content)
        if not ext:
            raise OperationNotSupported("FileSeries supports only image files for now. "
                                        "Other file types will be implemented in future.")

        return base64_encode(file.content)

    def download(self, destination: Optional[str]):
        target_dir = self._get_destination(destination)
        item_count = self._backend.get_image_series_values(self._run_uuid, self._path, 0, 1).totalItemCount
        for i in range(0, item_count):
            self._backend.download_file_series_by_index(self._run_uuid, self._path, i, target_dir)

    def download_last(self, destination: Optional[str]):
        target_dir = self._get_destination(destination)
        item_count = self._backend.get_image_series_values(self._run_uuid, self._path, 0, 1).totalItemCount
        if item_count > 0:
            self._backend.download_file_series_by_index(self._run_uuid, self._path, item_count - 1, target_dir)
        else:
            raise ValueError("Unable to download last file - series is empty")

    def _get_destination(self, destination: Optional[str]):
        target_dir = destination
        if destination is None:
            target_dir = os.path.join('neptune', self._path[-1])
        pathlib.Path(os.path.abspath(target_dir)).mkdir(parents=True, exist_ok=True)
        return target_dir



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest
from datetime import datetime

from neptune.new import ANONYMOUS, init
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import MetadataInconsistency, InactiveRunException
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.series import FloatSeries, StringSeries


class TestRun(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def test_define(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/value", Float(5), wait=True)
        self.assertEqual(exp.get_structure()['some']['path']['value'].fetch(), 5)

    def test_define_string(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/value", String("Some string"), wait=True)
        self.assertEqual(exp.get_structure()['some']['path']['value'].fetch(), "Some string")

    def test_define_few_variables(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/num", Float(3))
        exp.define("some/path/text", String("Some text"), wait=True)
        self.assertEqual(exp.get_structure()['some']['path']['num'].fetch(), 3)
        self.assertEqual(exp.get_structure()['some']['path']['text'].fetch(), "Some text")

    def test_define_conflict(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/value", Float(5))
        with self.assertRaises(MetadataInconsistency):
            exp.define("some/path/value", Float(1))

    def test_pop(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/num", Float(3))
        exp.define("some/path/text", String("Some text"))
        exp.pop("some/path/text")
        self.assertTrue('num' in exp.get_structure()['some']['path'])
        self.assertTrue('text' not in exp.get_structure()['some']['path'])

    def test_pop_namespace(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/subpath/num", Float(3))
        exp.define("some/path/text", String("Some text"))
        exp.define("some/otherpath", Float(4))
        exp.pop("some/path")
        self.assertTrue('path' not in exp.get_structure()['some'])

    def test_run_as_handler(self):
        exp = init(mode="debug", flush_period=0.5)
        exp.define("some/path/num", Float(3))
        exp.define("some/path/text", String("Some text"))
        handler = exp['some/path']
        exp.wait()
        self.assertEqual(handler['num'].fetch(), 3)
        self.assertEqual(handler['text'].fetch(), "Some text")

    def test_assign_dict(self):
        exp = init(mode="debug", flush_period=0.5)
        now = datetime.now()
        exp.assign({
            "x": 5,
            "metadata": {
                "name": "Trol",
                "age": 376
            },
            "toys": StringSeries(["cudgel", "hat"]),
            "nested": {
                "nested": {
                    "deep_secret": FloatSeries([13, 15])
                }
            },
            "simple_types": {
                "int": 42,
                "str": "imagine",
                "float": 3.14,
                "datetime": now,
                "list": list(range(10)),
            }
        })
        self.assertEqual(exp['x'].fetch(), 5)
        self.assertEqual(exp['metadata/name'].fetch(), "Trol")
        self.assertEqual(exp['metadata/age'].fetch(), 376)
        self.assertEqual(exp['toys'].fetch_last(), "hat")
        self.assertEqual(exp['nested/nested/deep_secret'].fetch_last(), 15)
        self.assertEqual(exp['simple_types/int'].fetch(), 42)
        self.assertEqual(exp['simple_types/str'].fetch(), 'imagine')
        self.assertEqual(exp['simple_types/float'].fetch(), 3.14)
        self.assertEqual(exp['simple_types/datetime'].fetch(),
                         now.replace(microsecond=1000 * int(now.microsecond / 1000)))
        self.assertEqual(exp['simple_types/list'].fetch(), str(list(range(10))))

    def test_assign_false(self):
        # https://github.com/neptune-ai/neptune-client/issues/555
        exp = init(mode="debug")
        exp["params"] = {'predictor.cheat': False}

        self.assertFalse(exp["params/predictor.cheat"].fetch())

    def test_access_blocked_after_stop(self):
        exp = init(mode="debug")
        exp['attr1'] = 1

        exp.stop()

        with self.assertRaises(InactiveRunException):
            exp['attr1'].fetch()
        with self.assertRaises(InactiveRunException):
            exp['attr2'] = 2
        with self.assertRaises(InactiveRunException):
            exp['series'].log(1)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
from io import StringIO

from neptune.backend import BackendApiClient, LeaderboardApiClient
from neptune.utils import NoopObject

_logger = logging.getLogger(__name__)


class OfflineBackendApiClient(BackendApiClient):
    def __init__(self):
        _logger.warning('Neptune is running in offline mode. No data is being logged to Neptune.')
        _logger.warning('Disable offline mode to log your experiments.')

    @property
    def api_address(self):
        return 'OFFLINE'

    @property
    def display_address(self):
        return 'OFFLINE'

    @property
    def proxies(self):
        return None

    def get_project(self, project_qualified_name):
        return NoopObject()

    def get_projects(self, namespace):
        return []

    def create_leaderboard_backend(self, project) -> 'OfflineLeaderboardApiClient':
        return OfflineLeaderboardApiClient()


class OfflineLeaderboardApiClient(LeaderboardApiClient):
    @property
    def api_address(self):
        return 'OFFLINE'

    @property
    def display_address(self):
        return 'OFFLINE'

    @property
    def proxies(self):
        return None

    def get_project_members(self, project_identifier):
        return []

    def get_leaderboard_entries(
            self, project, entry_types=None, ids=None, states=None, owners=None, tags=None, min_running_time=None):
        return []

    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        return StringIO()

    def get_metrics_csv(self, experiment):
        return StringIO()

    def create_experiment(self, project, name, description,
                          params, properties, tags, abortable,
                          monitored, git_info, hostname, entrypoint,
                          notebook_id, checkpoint_id):
        return NoopObject()

    def upload_source_code(self, experiment, source_target_pairs):
        pass

    def get_notebook(self, project, notebook_id):
        return NoopObject()

    def get_last_checkpoint(self, project, notebook_id):
        return NoopObject()

    def create_notebook(self, project):
        return NoopObject()

    def create_checkpoint(self, notebook_id, jupyter_path, _file=None):
        pass

    def get_experiment(self, experiment_id):
        return NoopObject()

    def update_experiment(self, experiment, properties):
        pass

    def set_property(self, experiment, key, value):
        pass

    def remove_property(self, experiment, key):
        pass

    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        pass

    def upload_experiment_source(self, experiment, data, progress_indicator):
        pass

    def extract_experiment_source(self, experiment, data):
        pass

    def create_channel(self, experiment, name, channel_type):
        return NoopObject()

    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        pass

    def get_channels(self, experiment):
        return {}

    def create_system_channel(self, experiment, name, channel_type):
        return NoopObject()

    def get_system_channels(self, experiment):
        return {}

    def send_channels_values(self, experiment, channels_with_values):
        pass

    def mark_succeeded(self, experiment):
        pass

    def mark_failed(self, experiment, traceback):
        pass

    def ping_experiment(self, experiment):
        pass

    def create_hardware_metric(self, experiment, metric):
        return NoopObject()

    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        pass

    def log_artifact(self, experiment, artifact, destination=None):
        pass

    def delete_artifacts(self, experiment, path):
        pass

    def rm_data(self, experiment, path):
        pass

    def download_data(self, experiment, path, destination):
        pass

    def download_sources(self, experiment, path=None, destination_dir=None):
        pass

    def download_artifacts(self, experiment, path=None, destination_dir=None):
        pass

    def download_artifact(self, experiment, path=None, destination_dir=None):
        pass


# define deprecated OfflineBackend class
OfflineBackend = OfflineBackendApiClient



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, List, TypeVar, Generic, Callable, Union

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.utils.paths import path_to_str

T = TypeVar('T')
Node = TypeVar('Node')


# pylint: disable=unused-argument
def _default_node_factory(path):
    return {}


class RunStructure(Generic[T, Node]):
    def __init__(self, node_factory: Optional[Callable[[List[str]], Node]] = None):
        if node_factory is None:
            node_factory = _default_node_factory

        self._structure = node_factory(path=[])
        self._node_factory = node_factory
        self._node_type = type(self._structure)

    def get_structure(self) -> Node:
        return self._structure

    def get(self, path: List[str]) -> Union[T, Node, None]:
        ref = self._structure

        for index, part in enumerate(path):
            if not isinstance(ref, self._node_type):
                raise MetadataInconsistency("Cannot access path '{}': '{}' is already defined as an attribute, "
                                            "not a namespace".format(path_to_str(path), path_to_str(path[:index])))
            if part not in ref:
                return None
            ref = ref[part]

        return ref

    def set(self, path: List[str], attr: T) -> None:
        ref = self._structure
        location, attribute_name = path[:-1], path[-1]

        for idx, part in enumerate(location):
            if part not in ref:
                ref[part] = self._node_factory(location[:idx + 1])
            ref = ref[part]
            if not isinstance(ref, self._node_type):
                raise MetadataInconsistency("Cannot access path '{}': '{}' is already defined as an attribute, "
                                            "not a namespace".format(path_to_str(path), part))

        if attribute_name in ref and isinstance(ref[attribute_name], self._node_type):
            if isinstance(attr, self._node_type):
                # in-between nodes are auto-created, so ignore it's OK unless we want to change the type
                return
            raise MetadataInconsistency("Cannot set attribute '{}'. It's a namespace".format(path_to_str(path)))

        ref[attribute_name] = attr

    def pop(self, path: List[str]) -> None:
        self._pop_impl(self._structure, path, path)

    def _pop_impl(self, ref, sub_path: List[str], attr_path: List[str]):
        if not sub_path:
            return

        head, tail = sub_path[0], sub_path[1:]
        if head not in ref:
            raise MetadataInconsistency("Cannot delete {}. Attribute not found.".format(path_to_str(attr_path)))

        if not tail:
            if isinstance(ref[head], self._node_type):
                raise MetadataInconsistency(
                    "Cannot delete {}. It's a namespace, not an attribute.".format(path_to_str(attr_path)))
            del ref[head]
        else:
            self._pop_impl(ref[head], tail, attr_path)
            if not ref[head]:
                del ref[head]

    def clear(self):
        self._structure.clear()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections.abc import MutableMapping
from typing import Any, Dict, TYPE_CHECKING, Iterator, List, Mapping, Union

from neptune.new.attributes.attribute import Attribute
from neptune.new.internal.run_structure import RunStructure
from neptune.new.internal.utils.generic_attribute_mapper import atomic_attribute_types_map, NoValue
from neptune.new.internal.utils.paths import parse_path, path_to_str
from neptune.new.types.namespace import Namespace as NamespaceVal

if TYPE_CHECKING:
    from neptune.new.run import Run


class Namespace(Attribute, MutableMapping):
    def __init__(self, run: 'Run', path: List[str]):
        Attribute.__init__(self, run, path)
        self._attributes = {}
        self._str_path = path_to_str(path)

    def __setitem__(self, k: str, v: Attribute) -> None:
        self._attributes[k] = v

    def __delitem__(self, k: str) -> None:
        del self._attributes[k]

    def __getitem__(self, k: str) -> Attribute:
        return self._attributes[k]

    def __len__(self) -> int:
        return len(self._attributes)

    def __iter__(self) -> Iterator[str]:
        yield from self._attributes.__iter__()

    def to_dict(self) -> Dict[str, Any]:
        result = {}
        for key, value in self._attributes.items():
            if isinstance(value, Namespace):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result

    def assign(self, value: Union[NamespaceVal, dict, Mapping], wait: bool = False):
        if not isinstance(value, NamespaceVal):
            value = NamespaceVal(value)

        for k, v in value.value.items():
            self._run[f"{self._str_path}/{k}"].assign(v, wait)

    def _collect_atom_values(self, attribute_dict) -> dict:
        result = {}
        for k, v in attribute_dict.items():
            if isinstance(v, dict):
                result[k] = self._collect_atom_values(v)
            else:
                attr_type, attr_value = v
                if attr_type in atomic_attribute_types_map and attr_value is not NoValue:
                    result[k] = v[1]
        return result

    def fetch(self) -> dict:
        attributes = self._backend.fetch_atom_attribute_values(self._run_uuid, self._path)
        run_struct = RunStructure()
        prefix_len = len(self._path)
        for attr_name, attr_type, attr_value in attributes:
            run_struct.set(parse_path(attr_name)[prefix_len:], (attr_type, attr_value))
        return self._collect_atom_values(run_struct.get_structure())


class NamespaceBuilder:
    def __init__(self, run: 'Run'):
        self._run = run

    def __call__(self, path: List[str]) -> Namespace:
        return Namespace(self._run, path)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import unittest
import uuid
from time import time

from neptune.new.exceptions import MetadataInconsistency

from neptune.new.internal.backends.api_model import DatetimeAttribute, FloatAttribute, StringAttribute, \
    FloatSeriesAttribute, StringSeriesAttribute, StringSetAttribute

from neptune.new.internal.operation import AssignFloat, AssignString, AssignDatetime, LogFloats, LogStrings, \
    AddStrings

from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock


class TestNeptuneBackendMock(unittest.TestCase):
    # pylint:disable=protected-access

    project_uuid = uuid.uuid4()

    def test_get_float_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignFloat(["x"], 5)])

        # when
        ret = backend.get_float_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(FloatAttribute(5), ret)

    def test_get_string_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abcx")])

        # when
        ret = backend.get_string_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(StringAttribute("abcx"), ret)

    def test_get_datetime_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        now = datetime.datetime.now()
        now = now.replace(microsecond=1000*int(now.microsecond/1000))
        backend.execute_operations(exp.uuid, [AssignDatetime(["x"], now)])

        # when
        ret = backend.get_datetime_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(DatetimeAttribute(now), ret)

    def test_get_float_series_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [LogFloats(["x"], [LogFloats.ValueType(5, None, time()),
                                                                LogFloats.ValueType(3, None, time())])])
        backend.execute_operations(exp.uuid, [LogFloats(["x"], [LogFloats.ValueType(2, None, time()),
                                                                LogFloats.ValueType(9, None, time())])])

        # when
        ret = backend.get_float_series_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(FloatSeriesAttribute(9), ret)

    def test_get_string_series_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [LogStrings(["x"], [LogStrings.ValueType("adf", None, time()),
                                                                 LogStrings.ValueType("sdg", None, time())])])
        backend.execute_operations(exp.uuid, [LogStrings(["x"], [LogStrings.ValueType("dfh", None, time()),
                                                                 LogStrings.ValueType("qwe", None, time())])])

        # when
        ret = backend.get_string_series_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(StringSeriesAttribute("qwe"), ret)

    def test_get_string_set_attribute(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AddStrings(["x"], {"abcx", "qwe"})])

        # when
        ret = backend.get_string_set_attribute(exp.uuid, ["x"])

        # then
        self.assertEqual(StringSetAttribute({"abcx", "qwe"}), ret)

    def test_get_float_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abc")])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_float_series_attribute(exp.uuid, ["x"])

    def test_get_string_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignFloat(["x"], 5)])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_string_attribute(exp.uuid, ["x"])

    def test_get_datetime_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abc")])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_datetime_attribute(exp.uuid, ["x"])

    def test_get_float_series_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abc")])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_float_series_attribute(exp.uuid, ["x"])

    def test_get_string_series_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abc")])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_string_series_attribute(exp.uuid, ["x"])

    def test_get_string_set_attribute_wrong_type(self):
        # given
        backend = NeptuneBackendMock()
        exp = backend.create_run(self.project_uuid)
        backend.execute_operations(exp.uuid, [AssignString(["x"], "abc")])

        # then
        with self.assertRaises(MetadataInconsistency):
            backend.get_string_set_attribute(exp.uuid, ["x"])



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import uuid
from datetime import datetime
from typing import List, Dict, Optional, Union

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.backends.api_model import LeaderboardEntry, AttributeWithProperties, AttributeType
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.utils.paths import join_paths, parse_path

logger = logging.getLogger(__name__)


class RunsTableEntry:

    def __init__(self, backend: HostedNeptuneBackend, _id: uuid.UUID, attributes: List[AttributeWithProperties]):
        self._backend = backend
        self._id = _id
        self._attributes = attributes

    def __getitem__(self, path: str) -> 'LeaderboardHandler':
        return LeaderboardHandler(self, path)

    def get_attribute_type(self, path: str) -> AttributeType:
        for attr in self._attributes:
            if attr.path == path:
                return attr.type
        raise ValueError("Could not find {} attribute".format(path))

    def get_attribute_value(self, path: str):
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.RUN_STATE:
                    return attr.properties.value
                if _type in (
                        AttributeType.FLOAT, AttributeType.INT, AttributeType.BOOL,
                        AttributeType.STRING, AttributeType.DATETIME,
                ):
                    return attr.properties.value
                if _type == AttributeType.FLOAT_SERIES or _type == AttributeType.STRING_SERIES:
                    return attr.properties.last
                if _type == AttributeType.IMAGE_SERIES:
                    raise MetadataInconsistency("Cannot get value for image series.")
                if _type == AttributeType.FILE:
                    raise MetadataInconsistency("Cannot get value for file attribute. Use download() instead.")
                if _type == AttributeType.FILE_SET:
                    raise MetadataInconsistency("Cannot get value for file set attribute. Use download() instead.")
                if _type == AttributeType.STRING_SET:
                    return set(attr.properties.values)
                if _type == AttributeType.GIT_REF:
                    return attr.properties.commit.commitId
                if _type == AttributeType.NOTEBOOK_REF:
                    return attr.properties.notebookName
                logger.error(
                    "Attribute type %s not supported in this version, yielding None. Recommended client upgrade.",
                    _type
                )
                return None
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_attribute(self, path: str, destination: Optional[str]):
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.FILE:
                    self._backend.download_file(self._id, parse_path(path), destination)
                    return
                raise MetadataInconsistency("Cannot download file from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))

    def download_file_set_attribute(self, path: str, destination: Optional[str]):
        for attr in self._attributes:
            if attr.path == path:
                _type = attr.type
                if _type == AttributeType.FILE_SET:
                    self._backend.download_file_set(self._id, parse_path(path), destination)
                    return
                raise MetadataInconsistency("Cannot download ZIP archive from attribute of type {}".format(_type))
        raise ValueError("Could not find {} attribute".format(path))


class LeaderboardHandler:

    def __init__(self, run: RunsTableEntry, path: str):
        self._run = run
        self._path = path

    def __getitem__(self, path: str) -> 'LeaderboardHandler':
        return LeaderboardHandler(self._run, join_paths(self._path, path))

    def get(self):
        return self._run.get_attribute_value(self._path)

    def download(self, destination: Optional[str]):
        attr_type = self._run.get_attribute_type(self._path)
        if attr_type == AttributeType.FILE:
            return self._run.download_file_attribute(self._path, destination)
        elif attr_type == AttributeType.FILE_SET:
            return self._run.download_file_set_attribute(self._path, destination)
        raise MetadataInconsistency("Cannot download file from attribute of type {}".format(attr_type))


class RunsTable:

    def __init__(self, backend: HostedNeptuneBackend, entries: List[LeaderboardEntry]):
        self._backend = backend
        self._entries = entries

    def to_runs(self) -> List[RunsTableEntry]:
        return [RunsTableEntry(self._backend, e.id, e.attributes) for e in self._entries]

    def to_pandas(self):
        # pylint:disable=import-outside-toplevel
        import pandas as pd

        def make_attribute_value(attribute: AttributeWithProperties) -> Optional[Union[str, float, datetime]]:
            _type = attribute.type
            _properties = attribute.properties
            if _type == AttributeType.RUN_STATE:
                return _properties.value
            if _type in (
                    AttributeType.FLOAT, AttributeType.INT, AttributeType.BOOL,
                    AttributeType.STRING, AttributeType.DATETIME,
            ):
                return _properties.value
            if _type == AttributeType.FLOAT_SERIES or _type == AttributeType.STRING_SERIES:
                return _properties.last
            if _type == AttributeType.IMAGE_SERIES:
                return None
            if _type == AttributeType.FILE or _type == AttributeType.FILE_SET:
                return None
            if _type == AttributeType.STRING_SET:
                return ",".join(_properties.values)
            if _type == AttributeType.GIT_REF:
                return _properties.commit.commitId
            if _type == AttributeType.NOTEBOOK_REF:
                return _properties.notebookName
            logger.error(
                "Attribute type %s not supported in this version, yielding None. Recommended client upgrade.",
                _type
            )
            return None

        def make_row(entry: LeaderboardEntry) -> Dict[str, Optional[Union[str, float, datetime]]]:
            row: Dict[str, Union[str, float, datetime]] = dict()
            for attr in entry.attributes:
                value = make_attribute_value(attr)
                if value is not None:
                    row[attr.path] = value
            return row

        def sort_key(attr):
            domain = attr.split('/')[0]
            if domain == 'sys':
                return 0, attr
            if domain == 'monitoring':
                return 2, attr
            return 1, attr

        rows = dict((n, make_row(entry)) for (n, entry) in enumerate(self._entries))

        df = pd.DataFrame.from_dict(data=rows, orient='index')
        df = df.reindex(sorted(df.columns, key=sort_key), axis='columns')
        return df



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from neptune.new.exceptions import InternalClientError
from neptune.new.internal.operation import AssignBool, AssignInt, Operation, AssignFloat, AssignString, LogFloats, \
    LogStrings, \
    ClearFloatLog, ClearStringLog, AddStrings, RemoveStrings, DeleteAttribute, ClearStringSet, LogImages, \
    ClearImageLog, UploadFile, AssignDatetime, ConfigFloatSeries, UploadFileSet, UploadFileContent, DeleteFiles
from neptune.new.internal.operation_visitor import OperationVisitor, Ret


class OperationApiObjectConverter(OperationVisitor[dict]):

    def convert(self, op: Operation) -> dict:
        return op.accept(self)

    def visit_assign_float(self, op: AssignFloat) -> dict:
        return {
            'value': op.value
        }

    def visit_assign_int(self, op: AssignInt) -> dict:
        return {
            'value': op.value
        }

    def visit_assign_bool(self, op: AssignBool) -> dict:
        return {
            'value': op.value
        }

    def visit_assign_string(self, op: AssignString) -> dict:
        return {
            'value': op.value
        }

    def visit_assign_datetime(self, op: AssignDatetime) -> Ret:
        return {
            'valueMilliseconds': int(1000 * op.value.timestamp())
        }

    def visit_upload_file(self, _: UploadFile) -> dict:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_content(self, _: UploadFileContent) -> dict:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_set(self, op: UploadFileSet) -> Ret:
        raise InternalClientError("Specialized endpoints should be used to upload file set attribute")

    def visit_log_floats(self, op: LogFloats) -> dict:
        return {
            'entries': [{
                'value': value.value,
                'step': value.step,
                'timestampMilliseconds': int(value.ts * 1000)
            } for value in op.values]
        }

    def visit_log_strings(self, op: LogStrings) -> dict:
        return {
            'entries': [{
                'value': value.value,
                'step': value.step,
                'timestampMilliseconds': int(value.ts * 1000)
            } for value in op.values]
        }

    def visit_log_images(self, op: LogImages) -> dict:
        return {
            'entries': [{
                'value': {
                    'data': value.value.data,
                    'name': value.value.name,
                    'description': value.value.description
                },
                'step': value.step,
                'timestampMilliseconds': int(value.ts * 1000)
            } for value in op.values]
        }

    def visit_clear_float_log(self, _: ClearFloatLog) -> dict:
        return {}

    def visit_clear_string_log(self, _: ClearStringLog) -> dict:
        return {}

    def visit_clear_image_log(self, _: ClearImageLog) -> dict:
        return {}

    def visit_config_float_series(self, op: ConfigFloatSeries) -> dict:
        return {
            'min': op.min,
            'max': op.max,
            'unit': op.unit
        }

    def visit_add_strings(self, op: AddStrings) -> dict:
        return {
            'values': list(op.values)
        }

    def visit_remove_strings(self, op: RemoveStrings) -> dict:
        return {
            'values': list(op.values)
        }

    def visit_delete_attribute(self, _: DeleteAttribute) -> dict:
        return {}

    def visit_clear_string_set(self, _: ClearStringSet) -> dict:
        return {}

    def visit_delete_files(self, op: DeleteFiles) -> Ret:
        return {
            'filePaths': list(op.file_paths)
        }



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
from collections import deque
from io import StringIO
from json import JSONDecodeError
from typing import Optional


class JsonFileSplitter:

    BUFFER_SIZE = 64 * 1024
    MAX_PART_READ = 8 * 1024

    def __init__(self, file_path: str):
        self._file = open(file_path, "r")
        self._decoder = json.JSONDecoder(strict=False)
        self._part_buffer = StringIO()
        self._parsed_queue = deque()
        self._start_pos = 0

    def close(self) -> None:
        self._file.close()
        self._part_buffer.close()

    def get(self) -> Optional[dict]:
        if self._parsed_queue:
            return self._parsed_queue.popleft()
        self._read_data()
        if self._parsed_queue:
            return self._parsed_queue.popleft()

    def _read_data(self):
        if self._part_buffer.tell() < self.MAX_PART_READ:
            data = self._file.read(self.BUFFER_SIZE)
            if not data:
                return
            if self._part_buffer.tell() > 0:
                data = self._reset_part_buffer() + data
            self._decode(data)

        if not self._parsed_queue:
            data = self._file.read(self.BUFFER_SIZE)
            while data:
                self._part_buffer.write(data)
                data = self._file.read(self.BUFFER_SIZE)
            data = self._reset_part_buffer()
            self._decode(data)

    def _decode(self, data: str):
        start = self._json_start(data)
        while start is not None:
            try:
                json_data, start = self._decoder.raw_decode(data, start)
            except JSONDecodeError:
                self._part_buffer.write(data[start:])
                break
            else:
                self._parsed_queue.append(json_data)
                start = self._json_start(data, start)

    @staticmethod
    def _json_start(data: str, start: int = 0) -> Optional[int]:
        try:
            return data.index("{", start)
        except ValueError:
            return None

    def _reset_part_buffer(self) -> str:
        data = self._part_buffer.getvalue()
        self._part_buffer.close()
        self._part_buffer = StringIO()
        return data



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=redefined-outer-name

import os
import uuid
from random import randint

import pytest

import neptune.new.sync
from neptune.new.constants import OFFLINE_DIRECTORY
from neptune.new.exceptions import ProjectNotFound
from neptune.new.internal.backends.api_model import Project
from neptune.new.internal.containers.disk_queue import DiskQueue
from neptune.new.internal.operation import Operation
from neptune.new.internal.utils.sync_offset_file import SyncOffsetFile
from neptune.new.sync import (
    ApiRun,
    get_project,
    get_qualified_name,
    sync_all_runs,
    sync_selected_runs,
    synchronization_status,
)


def a_run():
    return ApiRun(uuid.uuid4(), 'EXP-{}'.format(randint(42, 12342)), 'org', 'proj', False)


def prepare_runs(path):
    unsync_exp = a_run()
    sync_exp = a_run()
    registered_runs = (unsync_exp, sync_exp)

    execution_id = "exec-0"

    for exp in registered_runs:
        exp_path = path / "async" / str(exp.uuid) / execution_id
        exp_path.mkdir(parents=True)
        queue = DiskQueue(exp_path, lambda x: x, lambda x: x)
        queue.put('op-0')
        queue.put('op-1')

    SyncOffsetFile(path / "async" / str(unsync_exp.uuid) / execution_id / "last_ack_version").write(1)
    SyncOffsetFile(path / "async" / str(unsync_exp.uuid) / execution_id / "last_put_version").write(2)

    SyncOffsetFile(path / "async" / str(sync_exp.uuid) / execution_id / "last_ack_version").write(2)
    SyncOffsetFile(path / "async" / str(sync_exp.uuid) / execution_id / "last_put_version").write(2)

    def get_run_impl(run_id):
        for run in registered_runs:
            if run_id in (str(run.uuid), get_qualified_name(run)):
                return run

    return unsync_exp, sync_exp, get_run_impl


def prepare_offline_run(path):
    offline_exp_uuid = str(uuid.uuid4())
    offline_exp_path = path / OFFLINE_DIRECTORY / offline_exp_uuid
    offline_exp_path.mkdir(parents=True)

    queue = DiskQueue(offline_exp_path, lambda x: x, lambda x: x)
    queue.put('op-0')
    queue.put('op-1')
    SyncOffsetFile(path / OFFLINE_DIRECTORY / offline_exp_uuid / "last_put_version").write(2)

    return offline_exp_uuid


def test_list_runs(tmp_path, mocker, capsys):
    # given
    unsync_exp, sync_exp, get_run_impl = prepare_runs(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)

    # and
    mocker.patch.object(neptune.new.sync, 'get_run', get_run_impl)
    mocker.patch.object(Operation, 'from_dict')

    # when
    synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ''
    assert 'Synchronized runs:\n- {}'.format(get_qualified_name(sync_exp)) in captured.out
    assert 'Unsynchronized runs:\n- {}'.format(get_qualified_name(unsync_exp)) in captured.out
    assert 'Unsynchronized offline runs:\n- offline/{}'.format(offline_exp_uuid) in captured.out


def test_list_runs_when_no_run(tmp_path, capsys):
    (tmp_path / "async").mkdir()
    # when
    with pytest.raises(SystemExit):
        synchronization_status(tmp_path)

    # then
    captured = capsys.readouterr()
    assert captured.err == ''
    assert 'There are no Neptune runs' in captured.out


def test_sync_all_runs(tmp_path, mocker, capsys):
    # given
    unsync_exp, sync_exp, get_run_impl = prepare_runs(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)
    registered_offline_run = a_run()

    # and
    mocker.patch.object(neptune.new.sync, 'get_run', get_run_impl)
    mocker.patch.object(neptune.new.sync, 'backend')
    mocker.patch.object(neptune.new.sync.backend, 'execute_operations')
    mocker.patch.object(neptune.new.sync.backend, 'get_project',
                        lambda _: Project(uuid.uuid4(), 'project', 'workspace'))
    mocker.patch.object(neptune.new.sync, 'register_offline_run', lambda _: registered_offline_run)
    mocker.patch.object(Operation, 'from_dict', lambda x: x)

    # when
    sync_all_runs(tmp_path, 'foo')

    # then
    captured = capsys.readouterr()
    assert captured.err == ''
    assert ('Offline run {} registered as {}'
            .format(offline_exp_uuid, get_qualified_name(registered_offline_run))) in captured.out
    assert 'Synchronising {}'.format(get_qualified_name(unsync_exp)) in captured.out
    assert 'Synchronization of run {} completed.'.format(get_qualified_name(unsync_exp)) in captured.out
    assert 'Synchronising {}'.format(get_qualified_name(sync_exp)) not in captured.out

    # and
    # pylint: disable=no-member
    neptune.new.sync.backend.execute_operations.has_calls([
        mocker.call(unsync_exp.uuid, ['op-1']),
        mocker.call(registered_offline_run.uuid, ['op-1'])
    ], any_order=True)


def test_sync_selected_runs(tmp_path, mocker, capsys):
    # given
    unsync_exp, sync_exp, get_run_impl = prepare_runs(tmp_path)
    offline_exp_uuid = prepare_offline_run(tmp_path)
    registered_offline_exp = a_run()

    def get_run_impl_(run_id: str):
        if run_id in (str(registered_offline_exp.uuid), get_qualified_name(registered_offline_exp)):
            return registered_offline_exp
        else:
            return get_run_impl(run_id)

    # and
    mocker.patch.object(neptune.new.sync, 'get_run', get_run_impl_)
    mocker.patch.object(neptune.new.sync, 'backend')
    mocker.patch.object(neptune.new.sync.backend, 'execute_operations')
    mocker.patch.object(neptune.new.sync.backend, 'get_project',
                        lambda _: Project(uuid.uuid4(), 'project', 'workspace'))
    mocker.patch.object(neptune.new.sync, 'register_offline_run', lambda _: registered_offline_exp)
    mocker.patch.object(Operation, 'from_dict', lambda x: x)

    # when
    sync_selected_runs(tmp_path, 'some-name', [get_qualified_name(sync_exp), 'offline/' + offline_exp_uuid])

    # then
    captured = capsys.readouterr()
    assert captured.err == ''
    assert 'Synchronising {}'.format(get_qualified_name(sync_exp)) in captured.out
    assert 'Synchronization of run {} completed.'.format(get_qualified_name(sync_exp)) in captured.out
    assert 'Synchronising {}'.format(get_qualified_name(registered_offline_exp)) in captured.out
    assert 'Synchronization of run {} completed.'.format(get_qualified_name(registered_offline_exp)) \
           in captured.out
    assert 'Synchronising {}'.format(get_qualified_name(unsync_exp)) not in captured.out

    # and
    # pylint: disable=no-member
    neptune.new.sync.backend.execute_operations \
        .assert_called_with(registered_offline_exp.uuid, ['op-0', 'op-1'])


def test_get_project_no_name_set(mocker):
    # given
    mocker.patch.object(os, 'getenv')
    os.getenv.return_value = None

    # expect
    assert get_project(None) is None


def test_get_project_project_not_found(mocker):
    # given
    mocker.patch.object(neptune.new.sync, 'backend')
    mocker.patch.object(neptune.new.sync.backend, 'get_project')
    neptune.new.sync.backend.get_project.side_effect = ProjectNotFound('foo')

    # expect
    assert get_project('foo') is None


def test_sync_non_existent_run(tmp_path, mocker, capsys):
    # given
    mocker.patch.object(neptune.new.sync, 'get_project')
    mocker.patch.object(neptune.new.sync, 'get_run')
    neptune.new.sync.get_run.return_value = a_run()

    # when
    sync_selected_runs(tmp_path, 'foo', ['bar'])

    # then
    captured = capsys.readouterr()
    assert "Warning: Run 'bar' does not exist in location" in captured.err



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import ssl

from six.moves import urllib

from future.utils import PY3
from websocket import ABNF, create_connection


class WebsocketClientAdapter(object):
    def __init__(self):
        self._ws_client = None

    def connect(self, url, token, proxies=None):
        sslopt = None
        if os.getenv("NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"):
            sslopt = {"cert_reqs": ssl.CERT_NONE}

        proto = url.split(':')[0].replace('ws', 'http')
        proxy = proxies[proto] if proxies and proto in proxies else os.getenv("{}_PROXY".format(proto.upper()))

        if proxy:
            proxy_split = urllib.parse.urlparse(proxy).netloc.split(':')
            proxy_host = proxy_split[0]
            proxy_port = proxy_split[1] if len(proxy_split) > 1 else "80" if proto == "http" else "443"
        else:
            proxy_host = None
            proxy_port = None

        self._ws_client = create_connection(url, header=self._auth_header(token), sslopt=sslopt,
                                            http_proxy_host=proxy_host, http_proxy_port=proxy_port)

    def recv(self):
        if self._ws_client is None:
            raise WebsocketNotConnectedException()

        opcode, data = None, None

        while opcode != ABNF.OPCODE_TEXT:
            opcode, data = self._ws_client.recv_data()

        return data.decode("utf-8") if PY3 else data

    @property
    def connected(self):
        return self._ws_client and self._ws_client.connected

    def close(self):
        if self._ws_client:
            return self._ws_client.close()

    def abort(self):
        if self._ws_client:
            return self._ws_client.abort()

    def shutdown(self):
        if self._ws_client:
            return self._ws_client.shutdown()

    @classmethod
    def _auth_header(cls, token):
        return ["Authorization: Bearer " + token['access_token']]


class WebsocketNotConnectedException(Exception):
    def __init__(self):
        super(WebsocketNotConnectedException, self).__init__('Websocket client is not connected!')



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import random
import time
import unittest
import uuid
from typing import Optional

from mock import MagicMock
from neptune.new.internal.operation_processors.operation_processor import OperationProcessor

from neptune.new.internal.operation_processors.sync_operation_processor import SyncOperationProcessor

from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock

from neptune.new.run import Run


_now = time.time()


class TestAttributeBase(unittest.TestCase):

    @staticmethod
    def _create_run(processor: Optional[OperationProcessor] = None):
        backend = NeptuneBackendMock()
        exp = backend.create_run(uuid.uuid4())
        if processor is None:
            processor = SyncOperationProcessor(exp.uuid, backend)
        _run = Run(exp.uuid, backend, processor, MagicMock(), MagicMock(), MagicMock(), MagicMock())
        _run.sync()
        _run.start()
        return _run

    @staticmethod
    def _random_path():
        return ["some", "random", "path", str(uuid.uuid4())]

    @staticmethod
    def _random_wait():
        return bool(random.getrandbits(1))

    @staticmethod
    def _now():
        return _now



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
from collections import OrderedDict

from mock import MagicMock, patch

from neptune.projects import Project
from neptune.sessions import Session
from tests.neptune.api_objects_factory import a_project


@patch('neptune.internal.api_clients.hosted_api_clients.mixins.SwaggerClient.from_url',
       MagicMock())
@patch('neptune.internal.api_clients.hosted_api_clients.hosted_backend_api_client.NeptuneAuthenticator',
       MagicMock())
class TestSession(unittest.TestCase):

    # threading.RLock needs to be mocked, because it breaks the equality of Projects
    @patch('threading.RLock')
    def test_get_projects_with_given_namespace(self, _):
        # given
        api_projects = [a_project(), a_project()]

        # and
        backend = MagicMock()
        leaderboard = MagicMock()
        backend.get_projects.return_value = api_projects
        backend.create_leaderboard_backend.return_value = leaderboard

        # and
        session = Session(backend=backend)

        # and
        custom_namespace = 'custom_namespace'

        # when
        projects = session.get_projects(custom_namespace)

        # then
        expected_projects = OrderedDict(
            (custom_namespace + '/' + p.name, Project(leaderboard, p.id, custom_namespace, p.name))
            for p in api_projects
        )
        self.assertEqual(expected_projects, projects)

        # and
        backend.get_projects.assert_called_with(custom_namespace)


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import atexit
import threading
import time
import traceback
import uuid
from contextlib import AbstractContextManager
from datetime import datetime
from enum import Enum
from functools import wraps
from typing import Any, Dict, List, Optional, Union

import click

from neptune.exceptions import UNIX_STYLES
from neptune.new.attributes import create_attribute_from_type
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.namespace import NamespaceBuilder, Namespace as NamespaceAttr
from neptune.new.exceptions import (
    MetadataInconsistency, InactiveRunException, NeptunePossibleLegacyUsageException,
)
from neptune.new.handler import Handler
from neptune.new.internal.backends.api_model import AttributeType
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.operation import DeleteAttribute
from neptune.new.internal.operation_processors.operation_processor import OperationProcessor
from neptune.new.internal.run_structure import RunStructure
from neptune.new.internal.utils import (
    is_bool, is_float, is_float_like, is_int, is_string, is_string_like, verify_type, is_dict_like,
)
from neptune.new.internal.utils.paths import parse_path
from neptune.new.internal.value_to_attribute_visitor import ValueToAttributeVisitor
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.namespace import Namespace
from neptune.new.types.value import Value


class RunState(Enum):
    CREATED = 'created'
    STARTED = 'started'
    STOPPING = 'stopping'
    STOPPED = 'stopped'


def assure_run_not_stopped(fun):
    @wraps(fun)
    def inner_fun(self, *args, **kwargs):
        # pylint: disable=protected-access
        if self._state == RunState.STOPPED:
            raise InactiveRunException(short_id=self._short_id)
        return fun(self, *args, **kwargs)

    return inner_fun


LEGACY_METHODS = (
    'create_experiment',
    'send_metric', 'log_metric',
    'send_text', 'log_text',
    'send_image', 'log_image',
    'send_artifact', 'log_artifact', 'delete_artifacts',
    'download_artifact', 'download_sources', 'download_artifacts',
    'reset_log', 'get_parameters',
    'get_properties', 'set_property', 'remove_property',
    'get_hardware_utilization', 'get_numeric_channels_values',
)


class Run(AbstractContextManager):
    """A Run in Neptune is a representation of all metadata that you log to Neptune.

    Beginning when you start a tracked run with `neptune.init()` and ending when the script finishes
    or when you explicitly stop the experiment with `.stop()`.

    You can log many ML metadata types, including:
        * metrics
        * losses
        * model weights
        * images
        * interactive charts
        * predictions
        * and much more

    Examples:
        >>> import neptune.new as neptune

        >>> # Create new experiment
        ... run = neptune.init('my_workspace/my_project')

        >>> # Log parameters
        ... params = {'max_epochs': 10, 'optimizer': 'Adam'}
        ... run['parameters'] = params

        >>> # Log metadata
        ... run['train/metric_name'].log()
        >>> run['predictions'].log(image)
        >>> run['model'].upload(path_to_model)

        >>> # Log whatever else you want
        ... pass

        >>> # Stop tracking and clean up
        ... run.stop()

    You may also want to check `Run docs page`_.

    .. _Run docs page:
       https://docs.neptune.ai/api-reference/run
    """

    last_run = None  # "static" instance of recently created Run

    def __init__(
            self,
            _uuid: uuid.UUID,
            backend: NeptuneBackend,
            op_processor: OperationProcessor,
            background_job: BackgroundJob,
            workspace: str,
            project_name: str,
            short_id: str,
            monitoring_namespace: str = "monitoring",
    ):
        self._uuid = _uuid
        self._backend = backend
        self._op_processor = op_processor
        self._bg_job = background_job
        self._structure: RunStructure[Attribute, NamespaceAttr] = RunStructure(NamespaceBuilder(self))
        self._lock = threading.RLock()
        self._state = RunState.CREATED
        self._workspace = workspace
        self._project_name = project_name
        self._short_id = short_id
        self.monitoring_namespace = monitoring_namespace

        Run.last_run = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        traceback.print_exception(exc_type, exc_val, exc_tb)
        self.stop()

    def __getattr__(self, item):
        if item in LEGACY_METHODS:
            raise NeptunePossibleLegacyUsageException()
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")

    @assure_run_not_stopped
    def __getitem__(self, path: str) -> 'Handler':
        return Handler(self, path)

    @assure_run_not_stopped
    def __setitem__(self, key: str, value) -> None:
        self.__getitem__(key).assign(value)

    @assure_run_not_stopped
    def __delitem__(self, path) -> None:
        self.pop(path)

    @assure_run_not_stopped
    def assign(self, value, wait: bool = False) -> None:
        """Assign values to multiple fields from a dictionary.
        You can use this method to quickly log all run's parameters.

        Args:
            value (dict): A dictionary with values to assign, where keys become the paths of the fields.
                The dictionary can be nested - in such case the path will be a combination of all keys.
            wait (bool, optional): If `True` the client will first wait to send all tracked metadata to the server.
                This makes the call synchronous. Defaults to `True`.

        Examples:
            >>> import neptune.new as neptune
            >>> run = neptune.init()

            >>> # Assign multiple fields from a dictionary
            ... params = {"max_epochs": 10, "optimizer": "Adam"}
            >>> run["parameters"] = params

            >>> # You can always log explicitely parameters one by one
            ... run["parameters/max_epochs"] = 10
            >>> run["parameters/optimizer"] = "Adam"

            >>> # Dictionaries can be nested
            ... params = {"train": {"max_epochs": 10}}
            >>> run["parameters"] = params
            >>> # This will log 10 under path "parameters/train/max_epochs"

        You may also want to check `assign docs page`_.

        .. _assign docs page:
            https://docs.neptune.ai/api-reference/run#assign
        """
        self._get_root_handler().assign(value, wait)

    @assure_run_not_stopped
    def fetch(self) -> dict:
        """Fetch values of all non-File Atom fields as a dictionary.
        The result will preserve the hierarchical structure of the run's metadata, but will contain only non-File Atom
        fields.
        You can use this method to quickly retrieve previous run's parameters.

        Returns:
            `dict` containing all non-File Atom fields values.

        Examples:
            >>> import neptune.new as neptune
            >>> resumed_run = neptune.init(run="HEL-3")
            >>> params = resumed_run['model/parameters'].fetch()

            >>> run_data = resumed_run.fetch()

            >>> print(run_data)
            >>> # this will print out all Atom attributes stored in run as a dict

        You may also want to check `fetch docs page`_.

        .. _fetch docs page:
            https://docs.neptune.ai/api-reference/run#fetch
        """
        return self._get_root_handler().fetch()

    def ping(self):
        self._backend.ping_run(self._uuid)

    def start(self):
        atexit.register(self._shutdown_hook)
        self._op_processor.start()
        self._bg_job.start(self)
        self._state = RunState.STARTED

    def stop(self, seconds: Optional[Union[float, int]] = None) -> None:
        """Stops the tracked run and kills the synchronization thread.

        `.stop()` will be automatically called when a script that created the run finishes or on the destruction
        of Neptune context.

        When using Neptune with Jupyter notebooks it's a good practice to stop the tracked run manually as it
        will be stopped automatically only when the Jupyter kernel stops.

        Args:
            seconds (int or float, optional): Seconds to wait for all tracking calls to finish
                before stopping the tracked run.
                If `None` will wait for all tracking calls to finish. Defaults to `True`.

        Examples:
            If you are creating tracked runs from the script you don't need to call `.stop()`:

            >>> import neptune.new as neptune
            >>> run = neptune.init()

            >>> # Your training or monitoring code
            ... pass
            ... # If you are executing Python script .stop()
            ... # is automatically called at the end for every run

            If you are performing multiple training jobs from one script one after the other it is a good practice
            to `.stop()` the finished tracked runs as every open run keeps an open connection with Neptune,
            monitors hardware usage, etc. You can also use Context Managers - Neptune will automatically call `.stop()`
            on the destruction of Run context:

            >>> import neptune.new as neptune

            >>> # If you are running consecutive training jobs from the same script
            ... # stop the tracked runs manually at the end of single training job
            ... for config in configs:
            ...   run = neptune.init()
            ...   # Your training or monitoring code
            ...   pass
            ...   run.stop()

            >>> # You can also use with statement and context manager
            ... for config in configs:
            ...   with neptune.init() as run:
            ...     # Your training or monitoring code
            ...     pass
            ...     # .stop() is automatically called
            ...     # when code execution exits the with statement

        .. warning::
            If you are using Jupyter notebooks for creating your runs you need to manually invoke `.stop()` once the
            training and evaluation is done.

        You may also want to check `stop docs page`_.

        .. _stop docs page:
            https://docs.neptune.ai/api-reference/run#stop
        """
        verify_type("seconds", seconds, (float, int, type(None)))
        if self._state != RunState.STARTED:
            return

        self._state = RunState.STOPPING
        ts = time.time()
        click.echo("Shutting down background jobs, please wait a moment...")
        self._bg_job.stop()
        self._bg_job.join(seconds)
        click.echo("Done!")
        with self._lock:
            sec_left = None if seconds is None else seconds - (time.time() - ts)
            self._op_processor.stop(sec_left)
        self._state = RunState.STOPPED

    def get_structure(self) -> Dict[str, Any]:
        """Returns a run's metadata structure in form of a dictionary.

        This method can be used to traverse the run's metadata structure programmatically when using Neptune in
        automated workflows.

        .. danger::
            The returned object is a deep copy of an internal run's structure.

        Returns:
            ``dict``: with the run's metadata structure.

        """

        # This is very weird pylint false-positive.
        # pylint: disable=no-member
        return self._structure.get_structure().to_dict()

    def print_structure(self) -> None:
        """Pretty prints the structure of the run's metadata.

        Paths are ordered lexicographically and the whole structure is neatly colored.
        """
        self._print_structure_impl(self.get_structure(), indent=0)

    def get_run_url(self) -> str:
        """Returns the URL the run can be accessed with in the browser
        """
        return self._backend.get_run_url(self._uuid, self._workspace, self._project_name, self._short_id)

    def _print_structure_impl(self, struct: dict, indent: int) -> None:
        for key in sorted(struct.keys()):
            click.echo("    " * indent, nl=False)
            if isinstance(struct[key], dict):
                click.echo("{blue}'{key}'{end}:".format(
                    blue=UNIX_STYLES['blue'],
                    key=key,
                    end=UNIX_STYLES['end']))
                self._print_structure_impl(struct[key], indent=indent + 1)
            else:
                click.echo("{blue}'{key}'{end}: {type}".format(
                    blue=UNIX_STYLES['blue'],
                    key=key,
                    end=UNIX_STYLES['end'],
                    type=type(struct[key]).__name__))

    def define(self,
               path: str,
               value: Union[Value, int, float, str, datetime],
               wait: bool = False
               ) -> Attribute:
        if isinstance(value, Value):
            pass
        elif is_bool(value):
            value = Boolean(value)
        elif is_int(value):
            value = Integer(value)
        elif is_float(value):
            value = Float(value)
        elif is_string(value):
            value = String(value)
        elif isinstance(value, datetime):
            value = Datetime(value)
        elif is_float_like(value):
            value = Float(float(value))
        elif is_dict_like(value):
            value = Namespace(value)
        elif is_string_like(value):
            value = String(str(value))
        else:
            raise TypeError("Value of unsupported type {}".format(type(value)))
        parsed_path = parse_path(path)

        with self._lock:
            old_attr = self._structure.get(parsed_path)
            if old_attr:
                raise MetadataInconsistency("Attribute or namespace {} is already defined".format(path))
            attr = ValueToAttributeVisitor(self, parsed_path).visit(value)
            attr.assign(value, wait)
            self._structure.set(parsed_path, attr)
            return attr

    def get_attribute(self, path: str) -> Optional[Attribute]:
        with self._lock:
            return self._structure.get(parse_path(path))

    def set_attribute(self, path: str, attribute: Attribute) -> Optional[Attribute]:
        with self._lock:
            return self._structure.set(parse_path(path), attribute)

    def exists(self, path: str) -> bool:
        verify_type("path", path, str)
        return self.get_attribute(path) is not None

    def pop(self, path: str, wait: bool = False) -> None:
        """Removes the field stored under the path completely and all data associated with it.

        Args:
            path (str): Path of the field to be removed.
            wait (bool, optional): If `True` the client will first wait to send all tracked metadata to the server.
                This makes the call synchronous. Defaults to `True`.

        Examples:
            >>> import neptune.new as neptune
            >>> run = neptune.init()

            >>> run['parameters/learninggg_rata'] = 0.3

            >>> # Delete a field along with it's data
            ... run.pop('parameters/learninggg_rata')

            >>> run['parameters/learning_rate'] = 0.3

            >>> # Training finished
            ... run['trained_model'].upload('model.pt')
            >>> # 'model_checkpoint' is a File field
            ... run.pop('model_checkpoint')

        You may also want to check `pop docs page`_.

        .. _pop docs page:
           https://docs.neptune.ai/api-reference/run#pop
        """
        verify_type("path", path, str)
        with self._lock:
            self._pop_impl(parse_path(path), wait)

    def _pop_impl(self, parsed_path: List[str], wait: bool):
        attribute = self._structure.get(parsed_path)
        if isinstance(attribute, NamespaceAttr):
            self._pop_namespace(attribute, wait)
        else:
            self._op_processor.enqueue_operation(DeleteAttribute(parsed_path), wait)
            self._structure.pop(parsed_path)

    def _pop_namespace(self, namespace: NamespaceAttr, wait: bool):
        children = list(namespace)
        for key in children:
            sub_attr_path = namespace._path + [key] # pylint: disable=protected-access
            self._pop_impl(sub_attr_path, wait)

    def lock(self) -> threading.RLock:
        return self._lock

    def wait(self, disk_only=False) -> None:
        """Wait for all the tracking calls to finish.

        Args:
            disk_only (bool, optional, default is False): If `True` the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.
                Defaults to `False`.

        You may also want to check `wait docs page`_.

        .. _wait docs page:
            https://docs.neptune.ai/api-reference/run#wait
        """
        with self._lock:
            if disk_only:
                self._op_processor.flush()
            else:
                self._op_processor.wait()

    def sync(self, wait: bool = True) -> None:
        """Synchronizes local representation of the run with Neptune servers.

        Args:
            wait (bool, optional, default is False): If `True` the process will only wait for data to be saved
                locally from memory, but will not wait for them to reach Neptune servers.
                Defaults to `True`.

        Examples:
            >>> import neptune.new as neptune

            >>> # Connect to a run from Worker #3
            ... worker_id = 3
            >>> run = neptune.init(run='DIST-43', monitoring_namespace='monitoring/{}'.format(worker_id))

            >>> # Try to access logs that were created in meantime by Worker #2
            ... worker_2_status = run['status/2'].fetch() # Error if this field was created after this script starts

            >>> run.sync() # Synchronizes local representation with Neptune servers.
            >>> worker_2_status = run['status/2'].fetch() # No error

        You may also want to check `sync docs page`_.

        .. _sync docs page:
            https://docs.neptune.ai/api-reference/run#sync
        """
        with self._lock:
            if wait:
                self._op_processor.wait()
            attributes = self._backend.get_attributes(self._uuid)
            self._structure.clear()
            for attribute in attributes:
                self._define_attribute(parse_path(attribute.path), attribute.type)

    def _define_attribute(self, _path: List[str], _type: AttributeType):
        attr = create_attribute_from_type(_type, self, _path)
        self._structure.set(_path, attr)

    def _get_root_handler(self):
        return Handler(self, "")

    def _shutdown_hook(self):
        self.stop()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class MetricService(object):
    def __init__(self, backend, metric_reporter, experiment, metrics_container):
        self.__backend = backend
        self.__metric_reporter = metric_reporter
        self.experiment = experiment
        self.metrics_container = metrics_container

    def report_and_send(self, timestamp):
        metric_reports = self.__metric_reporter.report(timestamp)
        self.__backend.send_hardware_metric_reports(self.experiment, self.metrics_container.metrics(), metric_reports)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
from collections import OrderedDict

from neptune.internal.api_clients import HostedNeptuneBackendApiClient
from neptune.projects import Project
from neptune.utils import assure_project_qualified_name

_logger = logging.getLogger(__name__)


class Session(object):
    """A class for running communication with Neptune.

    In order to query Neptune experiments you need to instantiate this object first.

    Args:
        backend (:class:`~neptune.backend.ApiClient`, optional, default is ``None``):
            By default, Neptune client library sends logs, metrics, images, etc to Neptune servers:
            either publicly available SaaS, or an on-premises installation.

            You can pass the default backend instance explicitly to specify its parameters:

            .. code :: python3

                from neptune import Session, HostedNeptuneBackendApiClient
                session = Session(backend=HostedNeptuneBackendApiClient(...))

            Passing an instance of :class:`~neptune.OfflineApiClient` makes your code run without communicating
            with Neptune servers.

            .. code :: python3

                from neptune import Session, OfflineApiClient
                session = Session(backend=OfflineApiClient())

        api_token (:obj:`str`, optional, default is ``None``):
            User's API token. If ``None``, the value of ``NEPTUNE_API_TOKEN`` environment variable will be taken.
            Parameter is ignored if ``backend`` is passed.

            .. deprecated :: 0.4.4

            Instead, use:

            .. code :: python3

                from neptune import Session
                session = Session.with_default_backend(api_token='...')

        proxies (:obj:`str`, optional, default is ``None``):
            Argument passed to HTTP calls made via the `Requests <https://2.python-requests.org/en/master/>`_ library.
            For more information see their proxies
            `section <https://2.python-requests.org/en/master/user/advanced/#proxies>`_.
            Parameter is ignored if ``backend`` is passed.

            .. deprecated :: 0.4.4

            Instead, use:

            .. code :: python3

                from neptune import Session, HostedNeptuneBackendApiClient
                session = Session(backend=HostedNeptuneBackendApiClient(proxies=...))

    Examples:

        Create session, assuming you have created an environment variable ``NEPTUNE_API_TOKEN``

        .. code:: python3

            from neptune import Session
            session = Session.with_default_backend()

        Create session and pass ``api_token``

        .. code:: python3

            from neptune import Session
            session = Session.with_default_backend(api_token='...')

        Create an offline session

        .. code:: python3

            from neptune import Session, OfflineApiClient
            session = Session(backend=OfflineApiClient())

    """

    def __init__(self, api_token=None, proxies=None, backend=None):
        self._backend = backend

        if self._backend is None:
            _logger.warning('WARNING: Instantiating Session without specifying a backend is deprecated '
                            'and will be removed in future versions. For current behaviour '
                            'use `neptune.init(...)` or `Session.with_default_backend(...)')

            self._backend = HostedNeptuneBackendApiClient(api_token, proxies)

    @classmethod
    def with_default_backend(cls, api_token=None, proxies=None):
        """The simplest way to instantiate a ``Session``.

        Args:
            api_token (:obj:`str`):
                User's API token.
                If ``None``, the value of ``NEPTUNE_API_TOKEN`` environment variable will be taken.

            proxies (:obj:`str`, optional, default is ``None``):
                Argument passed to HTTP calls made via the `Requests <https://2.python-requests.org/en/master/>`_
                library.
                For more information see their proxies
                `section <https://2.python-requests.org/en/master/user/advanced/#proxies>`_.

        Examples:

            .. code :: python3

                from neptune import Session
                session = Session.with_default_backend()

        """
        return cls(backend=HostedNeptuneBackendApiClient(api_token=api_token,
                                                         proxies=proxies)
                   )

    def get_project(self, project_qualified_name):
        """Get a project with given ``project_qualified_name``.

        In order to access experiments data one needs to get a :class:`~neptune.projects.Project` object first.
        This method gives you the ability to do that.

        Args:
            project_qualified_name (:obj:`str`):
                Qualified name of a project in a form of ``namespace/project_name``.
                If ``None``, the value of ``NEPTUNE_PROJECT`` environment variable will be taken.

        Returns:
            :class:`~neptune.projects.Project` object.

        Raise:
            :class:`~neptune.api_exceptions.ProjectNotFound`: When a project with given name does not exist.

        Examples:

            .. code:: python3

                # Create a Session instance
                from neptune.sessions import Session
                session = Session()

                # Get a project by it's ``project_qualified_name``:
                my_project = session.get_project('namespace/project_name')

        """
        project_qualified_name = assure_project_qualified_name(project_qualified_name)

        return self._backend.get_project(project_qualified_name)

    def get_projects(self, namespace):
        """Get all projects that you have permissions to see in given workspace.

        | This method gets you all available projects names and their
          corresponding :class:`~neptune.projects.Project` objects.
        | Both private and public projects may be returned for the workspace.
          If you have role in private project, it is included.
        | You can retrieve all the public projects that belong to any user or workspace,
          as long as you know their username or workspace name.

        Args:
            namespace (:obj:`str`): It can either be name of the workspace or username.

        Returns:
            :obj:`OrderedDict`
                | **keys** are ``project_qualified_name`` that is: *'workspace/project_name'*
                | **values** are corresponding :class:`~neptune.projects.Project` objects.

        Raises:
            `WorkspaceNotFound`: When the given workspace does not exist.

        Examples:

            .. code:: python3

                # create Session
                from neptune.sessions import Session
                session = Session()

                # Now, you can list all the projects available for a selected namespace.
                # You can use `YOUR_NAMESPACE` which is your workspace or user name.
                # You can also list public projects created in other workspaces.
                # For example you can use the `neptune-ai` namespace.

                session.get_projects('neptune-ai')

                # Example output:
                # OrderedDict([('neptune-ai/credit-default-prediction',
                #               Project(neptune-ai/credit-default-prediction)),
                #              ('neptune-ai/GStore-Customer-Revenue-Prediction',
                #               Project(neptune-ai/GStore-Customer-Revenue-Prediction)),
                #              ('neptune-ai/human-protein-atlas',
                #               Project(neptune-ai/human-protein-atlas)),
                #              ('neptune-ai/Ships',
                #               Project(neptune-ai/Ships)),
                #              ('neptune-ai/Mapping-Challenge',
                #               Project(neptune-ai/Mapping-Challenge))
                #              ])
        """

        projects = [Project(self._backend.create_leaderboard_backend(p), p.id, namespace, p.name)
                    for p in self._backend.get_projects(namespace)]
        return OrderedDict((p.full_id, p) for p in projects)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os
import sys
import textwrap
import time
import uuid
from pathlib import Path
from typing import Sequence, Iterable, List, Optional, Tuple, Any

import click

from neptune.new.constants import (
    ASYNC_DIRECTORY,
    NEPTUNE_RUNS_DIRECTORY,
    OFFLINE_DIRECTORY,
    OFFLINE_NAME_PREFIX,
)
from neptune.new.envs import NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, PROJECT_ENV_NAME
from neptune.new.exceptions import (
    CannotSynchronizeOfflineRunsWithoutProject,
    NeptuneConnectionLostException, NeptuneException,
    ProjectNotFound,
    RunNotFound,
)
from neptune.new.internal.backends.api_model import Project, ApiRun
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.containers.disk_queue import DiskQueue
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import Operation

#######################################################################################################################
# Run and Project utilities
#######################################################################################################################


# Set in CLI entry points block, patched in tests
backend: NeptuneBackend = None

retries_timeout = int(os.getenv(NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, "3600"))


def get_run(run_id: str) -> Optional[ApiRun]:
    try:
        return backend.get_run(run_id)
    except RunNotFound:
        return None
    except NeptuneException as e:
        click.echo('Exception while fetching run {}. Skipping run.'.format(run_id), err=True)
        logging.exception(e)
        return None


project_name_missing_message = (
    'Project name not provided. Could not synchronize offline runs. '
    'To synchronize offline run, specify the project name with the --project flag '
    'or by setting the {} environment variable.'.format(PROJECT_ENV_NAME))


def project_not_found_message(project_name: str) -> str:
    return ('Project {} not found. Could not synchronize offline runs. '.format(project_name) +
            'Please ensure you specified the correct project name with the --project flag ' +
            'or with the {} environment variable, or contact Neptune for support.'.format(PROJECT_ENV_NAME))


def get_project(project_name_flag: Optional[str]) -> Optional[Project]:
    project_name = project_name_flag or os.getenv(PROJECT_ENV_NAME)
    if not project_name:
        click.echo(textwrap.fill(project_name_missing_message), file=sys.stderr)
        return None
    try:
        return backend.get_project(project_name)
    except ProjectNotFound:
        click.echo(textwrap.fill(project_not_found_message(project_name)), file=sys.stderr)
        return None


def get_qualified_name(run: ApiRun) -> str:
    return "{}/{}/{}".format(run.workspace, run.project_name, run.short_id)


def is_valid_uuid(val: Any) -> bool:
    try:
        uuid.UUID(val)
        return True
    except ValueError:
        return False


#######################################################################################################################
# Listing runs to be synchronized
#######################################################################################################################

def is_run_synced(run_path: Path) -> bool:
    return all(is_execution_synced(execution_path) for execution_path in run_path.iterdir())


def is_execution_synced(execution_path: Path) -> bool:
    disk_queue = DiskQueue(execution_path, lambda x: x.to_dict(), Operation.from_dict)
    return disk_queue.is_empty()


def get_offline_runs_ids(base_path: Path) -> List[str]:
    result = []
    if not (base_path / OFFLINE_DIRECTORY).is_dir():
        return []
    for run_path in (base_path / OFFLINE_DIRECTORY).iterdir():
        if is_valid_uuid(run_path.name):
            result.append(run_path.name)
    return result


def partition_runs(base_path: Path) -> Tuple[List[ApiRun], List[ApiRun], int]:
    synced_runs_uuids = []
    unsynced_runs_uuids = []
    for run_path in (base_path / ASYNC_DIRECTORY).iterdir():
        if is_valid_uuid(run_path.name):
            run_uuid = run_path.name
            if is_run_synced(run_path):
                synced_runs_uuids.append(run_uuid)
            else:
                unsynced_runs_uuids.append(run_uuid)
    synced_runs = [run for run in map(get_run, synced_runs_uuids)]
    unsynced_runs = [run for run in map(get_run, unsynced_runs_uuids)]
    not_found = len([exp for exp in synced_runs + unsynced_runs if not exp or exp.trashed])
    synced_runs = [exp for exp in synced_runs if exp and not exp.trashed]
    unsynced_runs = [exp for exp in unsynced_runs if exp and not exp.trashed]

    return synced_runs, unsynced_runs, not_found


offline_run_explainer = '''
Runs which execute offline are not created on the server and they are not assigned to projects;
instead, they are identified by UUIDs like the ones above.
When synchronizing offline runs, please specify the workspace and project using the "--project"
flag. Alternatively, you can set the environment variable
{} to the target workspace/project. See the examples below.
'''.format(PROJECT_ENV_NAME)


def list_runs(base_path: Path, synced_runs: Sequence[ApiRun],
              unsynced_runs: Sequence[ApiRun], offline_runs_ids: Sequence[str]) \
        -> None:
    if not synced_runs and not unsynced_runs and not offline_runs_ids:
        click.echo('There are no Neptune runs in {}'.format(base_path))
        sys.exit(1)

    if unsynced_runs:
        click.echo('Unsynchronized runs:')
        for run in unsynced_runs:
            click.echo('- {}'.format(get_qualified_name(run)))

    if synced_runs:
        click.echo('Synchronized runs:')
        for run in synced_runs:
            click.echo('- {}'.format(get_qualified_name(run)))

    if offline_runs_ids:
        click.echo('Unsynchronized offline runs:')
        for run_id in offline_runs_ids:
            click.echo('- {}{}'.format(OFFLINE_NAME_PREFIX, run_id))
        click.echo()
        click.echo(textwrap.fill(offline_run_explainer, width=90))

    if not unsynced_runs:
        click.echo()
        click.echo('There are no unsynchronized runs in {}'.format(base_path))

    if not synced_runs:
        click.echo()
        click.echo('There are no synchronized runs in {}'.format(base_path))

    click.echo()
    click.echo('Please run with the `neptune sync --help` to see example commands.')


def synchronization_status(base_path: Path) -> None:
    synced_runs, unsynced_runs, not_found = partition_runs(base_path)
    if not_found > 0:
        click.echo(
            "WARNING: {} runs was skipped because they are in trash or do not exist anymore.".format(not_found),
            sys.stderr)
    offline_runs_ids = get_offline_runs_ids(base_path)
    list_runs(base_path, synced_runs, unsynced_runs, offline_runs_ids)


#######################################################################################################################
# Run synchronization
#######################################################################################################################


def sync_run(run_path: Path, qualified_run_name: str) -> None:
    run_uuid = uuid.UUID(run_path.name)
    click.echo('Synchronising {}'.format(qualified_run_name))
    for execution_path in run_path.iterdir():
        sync_execution(execution_path, run_uuid)
    click.echo('Synchronization of run {} completed.'.format(qualified_run_name))


def sync_execution(execution_path: Path, run_uuid: uuid.UUID) -> None:
    disk_queue = DiskQueue(execution_path, lambda x: x.to_dict(), Operation.from_dict)
    while True:
        batch, version = disk_queue.get_batch(1000)
        if not batch:
            break

        start_time = time.monotonic()
        while True:
            try:
                backend.execute_operations(run_uuid, batch)
                break
            except NeptuneConnectionLostException as ex:
                if time.monotonic() - start_time > retries_timeout:
                    raise ex
                click.echo("Experiencing connection interruptions. "
                           "Will try to reestablish communication with Neptune.",
                           sys.stderr)

        disk_queue.ack(version)


def sync_all_registered_runs(base_path: Path) -> None:
    for run_path in (base_path / ASYNC_DIRECTORY).iterdir():
        if is_valid_uuid(run_path.name) and not is_run_synced(run_path):
            run_uuid = run_path.name
            run = get_run(run_uuid)
            if run:
                sync_run(run_path, get_qualified_name(run))


def sync_selected_registered_runs(base_path: Path, qualified_runs_names: Sequence[str]) -> None:
    for name in qualified_runs_names:
        run = get_run(name)
        if run:
            run_path = base_path / ASYNC_DIRECTORY / str(run.uuid)
            if run_path.exists():
                sync_run(run_path, name)
            else:
                click.echo("Warning: Run '{}' does not exist in location {}".format(name, base_path),
                           file=sys.stderr)


def register_offline_run(project: Project) -> Optional[ApiRun]:
    try:
        return backend.create_run(project.uuid)
    except Exception as e:
        click.echo('Exception occurred while trying to create a run '
                   'on the Neptune server. Please try again later',
                   file=sys.stderr)
        logging.exception(e)
        return None


def move_offline_run(base_path: Path, offline_uuid: str, server_uuid: str) -> None:
    (base_path / ASYNC_DIRECTORY / server_uuid).mkdir(parents=True)
    (base_path / OFFLINE_DIRECTORY / offline_uuid).rename(base_path / ASYNC_DIRECTORY / server_uuid / "exec-0-offline")


def register_offline_runs(base_path: Path, project: Project,
                          offline_runs_ids: Iterable[str]) -> List[ApiRun]:
    result = []
    for run_uuid in offline_runs_ids:
        if (base_path / OFFLINE_DIRECTORY / run_uuid).is_dir():
            run = register_offline_run(project)
            if run:
                move_offline_run(base_path, offline_uuid=run_uuid, server_uuid=str(run.uuid))
                click.echo('Offline run {} registered as {}'
                           .format(run_uuid, get_qualified_name(run)))
                result.append(run)
        else:
            click.echo('Offline run with UUID {} not found on disk.'.format(run_uuid), err=True)
    return result


def is_offline_run_name(name: str) -> bool:
    return name.startswith(OFFLINE_NAME_PREFIX) and is_valid_uuid(name[len(OFFLINE_NAME_PREFIX):])


def sync_offline_runs(base_path: Path, project_name: Optional[str], offline_run_ids: Sequence[str]):
    if offline_run_ids:
        project = get_project(project_name)
        if not project:
            raise CannotSynchronizeOfflineRunsWithoutProject
        registered_runs = register_offline_runs(base_path, project, offline_run_ids)
        offline_runs_names = [get_qualified_name(exp) for exp in registered_runs]
        sync_selected_registered_runs(base_path, offline_runs_names)


def sync_selected_runs(base_path: Path, project_name: Optional[str],
                       runs_names: Sequence[str]) -> None:
    other_runs_names = [name for name in runs_names if not is_offline_run_name(name)]
    sync_selected_registered_runs(base_path, other_runs_names)

    offline_runs_ids = [name[len(OFFLINE_NAME_PREFIX):] for name in runs_names
                        if is_offline_run_name(name)]
    sync_offline_runs(base_path, project_name, offline_runs_ids)


def sync_all_runs(base_path: Path, project_name: Optional[str]) -> None:
    sync_all_registered_runs(base_path)

    offline_runs_ids = get_offline_runs_ids(base_path)
    sync_offline_runs(base_path, project_name, offline_runs_ids)


#######################################################################################################################
# Entrypoint for the CLI utility
#######################################################################################################################


# pylint: disable=unused-argument
def get_neptune_path(ctx, param, path: str) -> Path:
    # check if path exists and contains a '.neptune' folder
    path = Path(path)
    if (path / NEPTUNE_RUNS_DIRECTORY).is_dir():
        return path / NEPTUNE_RUNS_DIRECTORY
    elif path.name == NEPTUNE_RUNS_DIRECTORY and path.is_dir():
        return path
    else:
        raise click.BadParameter("Path {} does not contain a '{}' folder.".format(path, NEPTUNE_RUNS_DIRECTORY))


path_option = click.option('--path', type=click.Path(exists=True, file_okay=False, resolve_path=True),
                           default=Path.cwd(), callback=get_neptune_path, metavar='<location>',
                           help="path to a directory containing a '.neptune' folder with stored runs")


@click.command()
@path_option
def status(path: Path) -> None:
    """List synchronized and unsynchronized runs in the given directory. Trashed runs are not listed.

    Neptune stores run data on disk in '.neptune' directories. In case a run executes offline
    or network is unavailable as the run executes, run data can be synchronized
    with the server with this command line utility.

    Examples:

    \b
    # List synchronized and unsynchronized runs in the current directory
    neptune status

    \b
    # List synchronized and unsynchronized runs in directory "foo/bar" without actually syncing
    neptune status --path foo/bar
    """

    # pylint: disable=global-statement
    global backend
    backend = HostedNeptuneBackend(Credentials())

    synchronization_status(path)


@click.command()
@path_option
@click.option('--run', 'runs_names', multiple=True, metavar='<run-name>',
              help="run name (workspace/project/short-id or UUID for offline runs) to synchronize.")
@click.option('-p', '--project', 'project_name', multiple=False, metavar='project-name',
              help="project name (workspace/project) where offline runs will be sent")
def sync(path: Path, runs_names: List[str], project_name: Optional[str]):
    """Synchronizes runs with unsent data with the server.

    Neptune stores run data on disk in '.neptune' directories. In case a run executes offline
    or network is unavailable as the run executes, run data can be synchronized
    with the server with this command line utility.

    You can list unsynchronized runs with `neptune status`

    Examples:

    \b
    # Synchronize all runs in the current directory
    neptune sync

    \b
    # Synchronize all runs in the given path
    neptune sync --path foo/bar

    \b
    # Synchronize only runs "NPT-42" and "NPT-43" in "workspace/project" in the current directory
    neptune sync --run workspace/project/NPT-42 --run workspace/project/NPT-43

    \b
    # Synchronise all runs in the current directory, sending offline runs to project "workspace/project"
    neptune sync --project workspace/project

    \b
    # Synchronize only the offline run with UUID offline/a1561719-b425-4000-a65a-b5efb044d6bb
    # to project "workspace/project"
    neptune sync --project workspace/project --run offline/a1561719-b425-4000-a65a-b5efb044d6bb
    """

    # pylint: disable=global-statement
    global backend
    backend = HostedNeptuneBackend(Credentials())

    if runs_names:
        sync_selected_runs(path, project_name, runs_names)
    else:
        sync_all_runs(path, project_name)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import unittest
import uuid
from neptune.new.internal.operation import *


# pylint: disable=protected-access


class TestOperations(unittest.TestCase):

    def test_serialization_to_dict(self):
        classes = {cls.__name__ for cls in all_subclasses(Operation)}
        for obj in self._list_objects():
            if obj.__class__.__name__ in classes:
                classes.remove(obj.__class__.__name__)
            deserialized_obj = Operation.from_dict(json.loads(json.dumps(obj.to_dict())))
            self.assertEqual(obj.__dict__, deserialized_obj.__dict__)
        self.assertEqual(classes, set())

    @staticmethod
    def _list_objects():
        now = datetime.now()
        return [
            AssignFloat(TestOperations._random_path(), 5.),
            AssignInt(TestOperations._random_path(), 5),
            AssignBool(TestOperations._random_path(), True),
            AssignBool(TestOperations._random_path(), False),
            AssignString(TestOperations._random_path(), "a\rsdf\thr"),
            AssignDatetime(TestOperations._random_path(), now.replace(microsecond=1000*int(now.microsecond/1000))),
            UploadFile(TestOperations._random_path(), "f.txt", "file/path/f.txt"),
            UploadFileContent(TestOperations._random_path(), "stream.txt", "some base64"),
            UploadFileSet(TestOperations._random_path(), ["file/path/*.txt", "another/file/path/*.txt"], True),
            UploadFileSet(TestOperations._random_path(), ["file/path/*.txt", "another/file/path/*.txt"], False),
            DeleteFiles(TestOperations._random_path(), {"file/path/*.txt", "dir/path/"}),
            LogFloats(TestOperations._random_path(), [
                LogFloats.ValueType(5, 4, 500),
                LogFloats.ValueType(3, None, 1000),
                LogFloats.ValueType(10, 10, 1234)
            ]),
            LogStrings(TestOperations._random_path(), [
                LogStrings.ValueType("jetybv", 1, 5),
                LogStrings.ValueType("ghs\ner", 3, 123),
                LogStrings.ValueType("r", None, 1356),
                LogStrings.ValueType("ghsr", 13, 53682)
            ]),
            LogImages(TestOperations._random_path(), [
                LogImages.ValueType(ImageValue("base64_image_1", "name1", "description1"), None, 2),
                LogImages.ValueType(ImageValue("base64_image_2", "name2", "description2"), 0, 5),
            ]),
            ClearFloatLog(TestOperations._random_path()),
            ClearStringLog(TestOperations._random_path()),
            ClearImageLog(TestOperations._random_path()),
            ConfigFloatSeries(TestOperations._random_path(), min=11, max=600, unit="kg/h"),
            AddStrings(TestOperations._random_path(), {"asef", "asrge4"}),
            RemoveStrings(TestOperations._random_path(), {"a\ne", "aeg\t4ger", "agrg"}),
            ClearStringSet(TestOperations._random_path()),
            DeleteAttribute(TestOperations._random_path())
        ]

    @staticmethod
    def _random_path():
        return ["some", "random", "path", str(uuid.uuid4())]



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

import numpy as np
import pandas as pd
from pandas.util.testing import assert_frame_equal

from neptune.utils import map_keys, map_values, as_list, align_channels_on_x, get_channel_name_stems, \
    merge_dataframes
from tests.neptune.random_utils import sort_df_by_columns


class TestMapValues(unittest.TestCase):
    def test_empty_map(self):
        # when
        mapped_dict = map_values(times_2, {})

        # then
        self.assertEqual({}, mapped_dict)

    def test_non_empty_map(self):
        # when
        mapped_dict = map_values(times_2, {'a': 2, 'b': 3})

        # then
        self.assertEqual({'a': 4, 'b': 6}, mapped_dict)


class TestMapKeys(unittest.TestCase):
    def test_empty_map(self):
        # when
        mapped_dict = map_keys(times_2, {})

        # then
        self.assertEqual({}, mapped_dict)

    def test_non_empty_map(self):
        # when
        mapped_dict = map_keys(times_2, {2: 'a', 3: 'b'})

        # then
        self.assertEqual({4: 'a', 6: 'b'}, mapped_dict)


class TestAsList(unittest.TestCase):

    def test_none(self):
        # expect
        self.assertEqual(None, as_list(None))

    def test_scalar(self):
        # expect
        self.assertEqual([1], as_list(1))

    def test_list(self):
        # expect
        self.assertEqual([2], as_list([2]))

    def test_dict(self):
        self.assertEqual([{'a': 1}], as_list({'a': 1}))


class TestAlignChannelsOnX(unittest.TestCase):

    def test_ordered_x(self):
        # when
        np.random.seed(1234)
        random_batch = np.random.random(10).tolist()
        random_epoch = np.random.random(5).tolist()
        random_odd = np.random.random(7).tolist()

        df = pd.DataFrame({'x_batch_channel': list(range(10)),
                           'y_batch_channel': random_batch,
                           'x_epoch_channel': list(range(5)) + [np.nan] * 5,
                           'y_epoch_channel': random_epoch + [np.nan] * 5,
                           'x_odd_channel': list(range(7)) + [np.nan] * 3,
                           'y_odd_channel': random_odd + [np.nan] * 3}, dtype=float)

        expected_result = pd.DataFrame({'x': list(range(10)),
                                        'batch_channel': random_batch,
                                        'epoch_channel': random_epoch + [np.nan] * 5,
                                        'odd_channel': random_odd + [np.nan] * 3}, dtype=float)
        expected_result = sort_df_by_columns(expected_result)

        # then
        result = align_channels_on_x(df)
        result = sort_df_by_columns(result)

        assert_frame_equal(result, expected_result)

    def test_shuffled_x(self):
        # when

        batch_x = [4, 2, 10, 28]
        epoch_x = [np.nan] + [1, 2, 21]
        odd_x = [21, 10, 15, 4]
        detached_x = [3, 5, 9] + [np.nan]

        batch_y = [7, 2, 9, 1]
        epoch_y = [np.nan, 3, 5, 9]
        odd_y = [21, 15, 4, 3]
        detached_y = [1, 5, 12, np.nan]

        df = pd.DataFrame({'x_batch_channel': batch_x, 'y_batch_channel': batch_y,
                           'x_epoch_channel': epoch_x, 'y_epoch_channel': epoch_y,
                           'x_odd_channel': odd_x, 'y_odd_channel': odd_y,
                           'x_detached_channel': detached_x, 'y_detached_channel': detached_y}, dtype=float)

        expected_result = pd.DataFrame({'x': [1, 2, 3, 4, 5, 9, 10, 15, 21, 28],
                                        'batch_channel': [np.nan, 2, np.nan, 7, np.nan, np.nan, 9, np.nan, np.nan, 1],
                                        'epoch_channel': [3, 5, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, 9,
                                                          np.nan],
                                        'odd_channel': [np.nan, np.nan, np.nan, 3, np.nan, np.nan, 15, 4, 21, np.nan],
                                        'detached_channel': [np.nan, np.nan, 1, np.nan, 5, 12, np.nan, np.nan, np.nan,
                                                             np.nan]}, dtype=float)
        expected_result = sort_df_by_columns(expected_result)

        # then
        result = align_channels_on_x(df)
        result = sort_df_by_columns(result)

        assert_frame_equal(result, expected_result)

    def test_fraction_x(self):
        # when

        batch_x = [1.2, 0.3, 0.9, 123.4]
        epoch_x = [np.nan] + [1.7, 2.9, 4.5]

        batch_y = [7.3, 2.1, 9.5, 1.2]
        epoch_y = [np.nan, 0.35, 5.4, 0.9]

        df = pd.DataFrame({'x_batch_channel': batch_x, 'y_batch_channel': batch_y,
                           'x_epoch_channel': epoch_x, 'y_epoch_channel': epoch_y}, dtype=float)

        expected_result = pd.DataFrame({'x': [0.3, 0.9, 1.2, 1.7, 2.9, 4.5, 123.4],
                                        'batch_channel': [2.1, 9.5, 7.3, np.nan, np.nan, np.nan, 1.2],
                                        'epoch_channel': [np.nan, np.nan, np.nan, 0.35, 5.4, 0.9, np.nan]}, dtype=float)
        expected_result = sort_df_by_columns(expected_result)

        # then
        result = align_channels_on_x(df)
        result = sort_df_by_columns(result)

        assert_frame_equal(result, expected_result)


class TestGetChannelNameStems(unittest.TestCase):

    def setUp(self):
        np.random.seed(1234)
        self.df = pd.DataFrame({'x_batch_channel': list(range(10)),
                                'y_batch_channel': np.random.random(10),
                                'x_epoch_channel': list(range(5)) + [np.nan] * 5,
                                'y_epoch_channel': np.random.random(10),
                                'x_odd_channel': list(range(7)) + [np.nan] * 3,
                                'y_odd_channel': np.random.random(10)})

    def test_names(self):
        correct_names = set(['epoch_channel', 'batch_channel', 'odd_channel'])
        self.assertEqual(set(get_channel_name_stems(self.df)), correct_names)


class TestMergeDataFrames(unittest.TestCase):

    def setUp(self):
        np.random.seed(1234)
        random_df1 = np.random.random(10).tolist()
        self.df1 = pd.DataFrame({'x': list(range(10)),
                                 'y1': random_df1})

        random_df2 = np.random.random(3).tolist()
        self.df2 = pd.DataFrame({'x': list(range(3)),
                                 'y2': random_df2})

        random_df3 = np.random.random(6).tolist()
        self.df3 = pd.DataFrame({'x': list(range(6)),
                                 'y3': random_df3})

        df_merged_outer = pd.DataFrame({'x': list(range(10)),
                                        'y1': random_df1,
                                        'y2': random_df2 + [np.nan] * 7,
                                        'y3': random_df3 + [np.nan] * 4})

        self.df_merged_outer = sort_df_by_columns(df_merged_outer)

    def test_merge_outer(self):
        result = merge_dataframes([self.df1, self.df2, self.df3], on='x', how='outer')
        result = sort_df_by_columns(result)
        assert_frame_equal(result, self.df_merged_outer)


class TestSortDfByColumns(unittest.TestCase):

    def test_letters_and_numbers(self):
        sorted_df = pd.DataFrame(columns=['1', '2', '3', 'a', 'b', 'c', 'd', ])
        shuffled_df = pd.DataFrame(columns=['c', 'a', '1', 'd', '3', '2', 'b'])

        assert_frame_equal(sort_df_by_columns(shuffled_df), sorted_df)


def times_2(x):
    return x * 2


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import unittest
# pylint: disable=protected-access
from dataclasses import dataclass
from datetime import datetime, timedelta
from io import StringIO, BytesIO
from tempfile import TemporaryDirectory, NamedTemporaryFile
from unittest.mock import patch

import PIL

from neptune.new.attributes.atoms.boolean import Boolean
from neptune.new.attributes.atoms.integer import Integer
from neptune.new.attributes.series import FileSeries

from neptune.new import init, ANONYMOUS
from neptune.new.attributes.atoms.datetime import Datetime
from neptune.new.attributes.atoms.file import File
from neptune.new.attributes.atoms.float import Float
from neptune.new.attributes.atoms.string import String
from neptune.new.attributes.sets.string_set import StringSet
from neptune.new.envs import PROJECT_ENV_NAME, API_TOKEN_ENV_NAME
from neptune.new.exceptions import FileNotFound, MissingFieldException
from neptune.new.types import File as FileVal
from neptune.new.types.atoms.datetime import Datetime as DatetimeVal
from neptune.new.types.atoms.float import Float as FloatVal
from neptune.new.types.atoms.string import String as StringVal
from neptune.new.types.namespace import Namespace as NamespaceVal
from neptune.new.types.series.file_series import FileSeries as FileSeriesVal
from neptune.new.types.series.float_series import FloatSeries as FloatSeriesVal
from neptune.new.types.series.string_series import StringSeries as StringSeriesVal
from neptune.new.types.sets.string_set import StringSet as StringSetVal


class TestHandler(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def test_set(self):
        exp = init(mode="debug", flush_period=0.5)
        now = datetime.now()
        exp['some/num/val'] = 5.
        exp['some/str/val'] = "some text"
        exp['some/datetime/val'] = now
        exp.wait()
        self.assertEqual(exp['some/num/val'].fetch(), 5.)
        self.assertEqual(exp['some/str/val'].fetch(), "some text")
        self.assertEqual(exp['some/datetime/val'].fetch(), now.replace(microsecond=1000*int(now.microsecond/1000)))
        self.assertIsInstance(exp.get_structure()['some']['num']['val'], Float)
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], String)
        self.assertIsInstance(exp.get_structure()['some']['datetime']['val'], Datetime)

    def test_assign_atom(self):
        exp = init(mode="debug", flush_period=0.5)
        now = datetime.now()
        exp['some/num/val'].assign(5.)
        exp['some/int/val'].assign(50)
        exp['some/str/val'].assign("some text", wait=True)
        exp['some/bool/val'].assign(True)
        exp['some/datetime/val'].assign(now)
        self.assertEqual(exp['some/num/val'].fetch(), 5.)
        self.assertEqual(exp['some/int/val'].fetch(), 50)
        self.assertEqual(exp['some/str/val'].fetch(), "some text")
        self.assertEqual(exp['some/bool/val'].fetch(), True)
        self.assertEqual(exp['some/datetime/val'].fetch(), now.replace(microsecond=1000*int(now.microsecond/1000)))
        self.assertIsInstance(exp.get_structure()['some']['num']['val'], Float)
        self.assertIsInstance(exp.get_structure()['some']['int']['val'], Integer)
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], String)
        self.assertIsInstance(exp.get_structure()['some']['bool']['val'], Boolean)
        self.assertIsInstance(exp.get_structure()['some']['datetime']['val'], Datetime)

        now = now + timedelta(seconds=3, microseconds=500000)
        exp['some/num/val'].assign(FloatVal(15))
        exp['some/str/val'].assign(StringVal("other text"), wait=False)
        exp['some/datetime/val'].assign(DatetimeVal(now), wait=True)
        self.assertEqual(exp['some/num/val'].fetch(), 15)
        self.assertEqual(exp['some/str/val'].fetch(), "other text")
        self.assertEqual(exp['some/datetime/val'].fetch(), now.replace(microsecond=1000*int(now.microsecond/1000)))
        self.assertIsInstance(exp.get_structure()['some']['num']['val'], Float)
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], String)
        self.assertIsInstance(exp.get_structure()['some']['datetime']['val'], Datetime)

    @patch('neptune.new.internal.backends.neptune_backend_mock.copyfile')
    def test_save_download_file(self, copy_mock):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/file'].upload("path/to/other/file.txt")
        self.assertIsInstance(exp.get_structure()['some']['num']['file'], File)

        exp['some/num/file'].download("path/to/download/alternative_file.txt")
        copy_mock.assert_called_with(
            os.path.abspath("path/to/other/file.txt"), os.path.abspath("path/to/download/alternative_file.txt"))
        copy_mock.reset_mock()

        exp['some/num/file'].download()
        copy_mock.assert_called_with(
            os.path.abspath("path/to/other/file.txt"), os.path.abspath("file.txt"))
        copy_mock.reset_mock()

        exp['some/num/file'].download("path/to/other/file.txt")
        copy_mock.assert_not_called()

        with self.assertRaises(TypeError):
            exp['some/num/file'].download(123)

    def test_save_download_text_stream_to_given_destination(self):
        exp = init(mode="debug", flush_period=0.5)
        data = "Some test content of the stream"

        exp['some/num/attr_name'] = FileVal.from_stream(StringIO(data))
        self.assertIsInstance(exp.get_structure()['some']['num']['attr_name'], File)

        with NamedTemporaryFile("w") as temp_file:
            exp['some/num/attr_name'].download(temp_file.name)
            with open(temp_file.name, "rt") as file:
                self.assertEqual(file.read(), data)

    def test_save_download_binary_stream_to_default_destination(self):
        exp = init(mode="debug", flush_period=0.5)
        data = b"Some test content of the stream"

        exp['some/num/attr_name'] = FileVal.from_stream(BytesIO(data))
        self.assertIsInstance(exp.get_structure()['some']['num']['attr_name'], File)

        with TemporaryDirectory() as temp_dir:
            with patch('neptune.new.internal.backends.neptune_backend_mock.os.path.abspath') as abspath_mock:
                abspath_mock.side_effect = lambda path: os.path.normpath(temp_dir + "/" + path)
                exp['some/num/attr_name'].download()
            with open(temp_dir + '/attr_name.bin', "rb") as file:
                self.assertEqual(file.read(), data)

    @patch('neptune.new.internal.utils.glob', new=lambda path, recursive=False: [path.replace('*', 'file.txt')])
    @patch('neptune.new.internal.backends.neptune_backend_mock.ZipFile.write')
    def test_save_files_download(self, zip_write_mock):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/artifacts'].upload_files("path/to/file.txt")
        exp['some/artifacts'].upload_files("path/to/other/*")

        with TemporaryDirectory() as temp_dir:
            exp['some/artifacts'].download(temp_dir)
            exp['some/artifacts'].download(temp_dir)

        zip_write_mock.assert_any_call(os.path.abspath("path/to/file.txt"), "path/to/file.txt")
        zip_write_mock.assert_any_call(os.path.abspath("path/to/other/file.txt"), "path/to/other/file.txt")

    def test_assign_series(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/val'].assign(FloatSeriesVal([1, 2, 0, 10]))
        exp['some/str/val'].assign(StringSeriesVal(["text1", "text2"]), wait=True)
        exp['some/img/val'].assign(FileSeriesVal([FileVal.as_image(PIL.Image.new('RGB', (10, 15), color='red'))]))
        self.assertEqual(exp['some']['num']['val'].fetch_last(), 10)
        self.assertEqual(exp['some']['str']['val'].fetch_last(), "text2")
        self.assertIsInstance(exp.get_structure()['some']['img']['val'], FileSeries)

        exp['some/num/val'].assign(FloatSeriesVal([122, 543, 2, 5]))
        exp['some/str/val'].assign(StringSeriesVal(["other 1", "other 2", "other 3"]), wait=True)
        self.assertEqual(exp['some']['num']['val'].fetch_last(), 5)
        self.assertEqual(exp['some']['str']['val'].fetch_last(), "other 3")

    def test_log(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/val'].log(5)
        exp['some/str/val'].log("some text")
        exp['some/img/val'].log(FileVal.as_image(PIL.Image.new('RGB', (60, 30), color='red')))
        exp['some/img/val'].log(PIL.Image.new('RGB', (60, 30), color='red'))
        self.assertEqual(exp['some']['num']['val'].fetch_last(), 5)
        self.assertEqual(exp['some']['str']['val'].fetch_last(), "some text")
        self.assertIsInstance(exp.get_structure()['some']['img']['val'], FileSeries)

    def test_log_many_values(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/val'].log([5, 10, 15])
        exp['some/str/val'].log(["some text", "other"])
        exp['some/img/val'].log([FileVal.as_image(PIL.Image.new('RGB', (60, 30), color='red')),
                                 FileVal.as_image(PIL.Image.new('RGB', (20, 90), color='red'))])
        self.assertEqual(exp['some']['num']['val'].fetch_last(), 15)
        self.assertEqual(exp['some']['str']['val'].fetch_last(), "other")
        self.assertIsInstance(exp.get_structure()['some']['img']['val'], FileSeries)

    def test_log_value_errors(self):
        exp = init(mode="debug", flush_period=0.5)
        img = FileVal.as_image(PIL.Image.new('RGB', (60, 30), color='red'))

        with self.assertRaises(ValueError):
            exp['x'].log([])
        with self.assertRaises(ValueError):
            exp['x'].log([5, "str"])
        with self.assertRaises(ValueError):
            exp['x'].log([5, 10], step=10)

        exp['some/num/val'].log([5], step=1)
        exp['some/num/val'].log([])
        with self.assertRaises(ValueError):
            exp['some/num/val'].log("str")
        with self.assertRaises(TypeError):
            exp['some/num/val'].log(img)

        exp['some/str/val'].log(["str"], step=1)
        exp['some/str/val'].log([])

        exp['some/img/val'].log([img], step=1)
        exp['some/img/val'].log([])
        with self.assertRaises(TypeError):
            exp['some/img/val'].log(5)
        with self.assertRaises(FileNotFound):
            exp['some/img/val'].log("path")

        self.assertEqual(exp['some']['num']['val'].fetch_last(), 5)
        self.assertEqual(exp['some']['str']['val'].fetch_last(), "str")
        self.assertIsInstance(exp.get_structure()['some']['img']['val'], FileSeries)

    def test_assign_set(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/str/val'].assign(StringSetVal(["tag1", "tag2"]), wait=True)
        self.assertEqual(exp['some/str/val'].fetch(), {"tag1", "tag2"})
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], StringSet)

        exp['some/str/val'].assign(StringSetVal(["other_1", "other_2", "other_3"]), wait=True)
        self.assertEqual(exp['some/str/val'].fetch(), {"other_1", "other_2", "other_3"})
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], StringSet)

    def test_assign_namespace(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/namespace'].assign(
            NamespaceVal(
                {
                    'sub-namespace/val1': 1.0,
                    'sub-namespace/val2': StringSetVal(["tag1", "tag2"])
                }
            )
        )
        self.assertEqual(exp['some/namespace/sub-namespace/val1'].fetch(), 1.0)
        self.assertEqual(exp['some/namespace/sub-namespace/val2'].fetch(), {"tag1", "tag2"})
        self.assertIsInstance(exp.get_structure()['some']['namespace']['sub-namespace']['val1'], Float)
        self.assertIsInstance(exp.get_structure()['some']['namespace']['sub-namespace']['val2'], StringSet)

        exp['some'].assign(NamespaceVal({'namespace/sub-namespace/val1': 2.0}))
        self.assertEqual(exp['some/namespace/sub-namespace/val1'].fetch(), 2.0)
        self.assertEqual(exp['some/namespace/sub-namespace/val2'].fetch(), {"tag1", "tag2"})
        self.assertIsInstance(exp.get_structure()['some']['namespace']['sub-namespace']['val1'], Float)
        self.assertIsInstance(exp.get_structure()['some']['namespace']['sub-namespace']['val2'], StringSet)

        with self.assertRaises(TypeError):
            exp['some'].assign(NamespaceVal({'namespace/sub-namespace/val1': {"tagA", "tagB"}}))

    def test_assign_distinct_types(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/str/val'].assign(FloatVal(1.0), wait=True)
        self.assertEqual(exp['some/str/val'].fetch(), 1.0)
        self.assertIsInstance(exp.get_structure()['some']['str']['val'], Float)

        with self.assertRaises(TypeError):
            exp['some/str/val'].assign(StringSetVal(["other_1", "other_2", "other_3"]), wait=True)

    def test_add(self):
        exp = init(mode="debug", flush_period=0.5)

        exp['some/str/val'].add(["some text", "something else"], wait=True)
        self.assertEqual(exp['some/str/val'].fetch(), {"some text", "something else"})

        exp['some/str/val'].add("one more", wait=True)
        self.assertEqual(exp['some/str/val'].fetch(), {"some text", "something else", "one more"})

        self.assertIsInstance(exp.get_structure()['some']['str']['val'], StringSet)

    def test_pop(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/val'].assign(3, wait=True)
        self.assertIn('some', exp.get_structure())
        ns = exp['some']
        ns.pop('num/val', wait=True)
        self.assertNotIn('some', exp.get_structure())

    def test_pop_self(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['x'].assign(3, wait=True)
        self.assertIn('x', exp.get_structure())
        exp['x'].pop(wait=True)
        self.assertNotIn('x', exp.get_structure())

    def test_del(self):
        exp = init(mode="debug", flush_period=0.5)
        exp['some/num/val'].assign(3)
        self.assertIn('some', exp.get_structure())
        ns = exp['some']
        del ns['num/val']
        self.assertNotIn('some', exp.get_structure())

    def test_lookup(self):
        exp = init(mode="debug", flush_period=0.5)
        ns = exp['some/ns']
        ns['val'] = 5
        exp.wait()
        self.assertEqual(exp['some/ns/val'].fetch(), 5)

        ns = exp['other/ns']
        exp['other/ns/some/value'] = 3
        exp.wait()
        self.assertEqual(ns['some/value'].fetch(), 3)

    def test_attribute_error(self):
        exp = init(mode="debug", flush_period=0.5)
        with self.assertRaises(MissingFieldException):
            exp['var'].something()
        with self.assertRaises(AttributeError):
            exp['var'].something()
        with self.assertRaises(KeyError):
            exp['var'].something()

    def test_float_like_types(self):
        exp = init(mode="debug", flush_period=0.5)

        exp.define("attr1", self.FloatLike(5))
        self.assertEqual(exp['attr1'].fetch(), 5)
        exp["attr1"] = "234"
        self.assertEqual(exp['attr1'].fetch(), 234)
        with self.assertRaises(ValueError):
            exp["attr1"] = "234a"

        exp["attr2"].assign(self.FloatLike(34))
        self.assertEqual(exp['attr2'].fetch(), 34)
        exp["attr2"].assign("555")
        self.assertEqual(exp['attr2'].fetch(), 555)
        with self.assertRaises(ValueError):
            exp["attr2"].assign("string")

        exp["attr3"].log(self.FloatLike(34))
        self.assertEqual(exp['attr3'].fetch_last(), 34)
        exp["attr3"].log(["345", self.FloatLike(34), 4, 13.])
        self.assertEqual(exp['attr3'].fetch_last(), 13)
        with self.assertRaises(ValueError):
            exp["attr3"].log([4, "234a"])

    def test_string_like_types(self):
        exp = init(mode="debug", flush_period=0.5)

        exp["attr1"] = "234"
        exp["attr1"] = self.FloatLike(12356)
        self.assertEqual(exp['attr1'].fetch(), "TestHandler.FloatLike(value=12356)")

        exp["attr2"].log("xxx")
        exp["attr2"].log(["345", self.FloatLike(34), 4, 13.])
        self.assertEqual(exp['attr2'].fetch_last(), "13.0")

    def test_assign_dict(self):
        exp = init(mode="debug", flush_period=0.5)
        exp["params"] = {
            "x": 5,
            "metadata": {
                "name": "Trol",
                "age": 376
            },
            "toys": StringSeriesVal(["cudgel", "hat"]),
            "nested": {
                "nested": {
                    "deep_secret": FloatSeriesVal([13, 15])
                }
            }
        }
        self.assertEqual(exp['params/x'].fetch(), 5)
        self.assertEqual(exp['params/metadata/name'].fetch(), "Trol")
        self.assertEqual(exp['params/metadata/age'].fetch(), 376)
        self.assertEqual(exp['params/toys'].fetch_last(), "hat")
        self.assertEqual(exp['params/nested/nested/deep_secret'].fetch_last(), 15)

    def test_fetch_dict(self):
        now = datetime.now()

        exp = init(mode='debug', flush_period=0.5)
        exp['params/int'] = 1
        exp['params/float'] = 3.14
        exp['params/bool'] = True
        exp['params/datetime'] = now
        exp['params/sub-namespace/int'] = 42
        exp['params/sub-namespace/string'] = 'Some text'

        # attributes to be ignored
        exp['params/sub-namespace/string_series'].log('Some text #1')
        exp['params/sub-namespace/int_series'].log(100)
        exp['some/num/attr_name'] = FileVal.from_stream(BytesIO(b'Some stream'))

        # pylint: disable=assignment-from-no-return
        # that's a false positive, pylint looks at Handler.fetch()
        params_dict = exp['params'].fetch()
        self.assertDictEqual(params_dict, {
            'int': 1,
            'float': 3.14,
            'bool': True,
            'datetime': now.replace(microsecond=1000 * int(now.microsecond / 1000)),
            'sub-namespace': {
                'int': 42,
                'string': 'Some text',
            }
        })

    def test_fetch_dict_with_path(self):
        now = datetime.now()

        exp = init(mode='debug', flush_period=0.5)
        exp['params/int'] = 1
        exp['params/float'] = 3.14
        exp['params/bool'] = True
        exp['params/datetime'] = now
        exp['params/sub-namespace/int'] = 42
        exp['params/sub-namespace/string'] = 'Some text'

        # pylint: disable=assignment-from-no-return
        # that's a false positive, pylint looks at Handler.fetch()
        params_dict = exp['params/sub-namespace'].fetch()
        self.assertDictEqual(params_dict, {
            'int': 42,
            'string': 'Some text',
        })

    @dataclass
    class FloatLike:

        value: float

        def __float__(self):
            return float(self.value)
